#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# virocon doc documentation build configuration file, created by
# sphinx-quickstart on Tue Aug 22 11:54:21 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from os import path
import codecs


sys.path.insert(0, os.path.abspath("../"))
here = os.path.abspath(os.path.dirname(__file__))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md', '.py']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "virocon"
copyright = "2021, virocon"
author = "virocon developers"


# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
def read(rel_path):
    here = path.abspath(path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


version = get_version("../virocon/__init__.py")

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "my_classic_theme"
html_theme_path = ["_theme"]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# black = "#000000"
# white = "#FFFFFF"
# grey = "#C0C0C0"
# light_blue = "#0040FF"
# blue = "#045FB4"
# light_grey = "#E6E6E6"
# dark_grey = "#424242"
# collapse_button = "#A4A4A4"
# html_theme_options = {
#     "rightsidebar" : "False", #Put the sidebar on the right side.
#     #"stickysidebar" : "True", #Make the sidebar “fixed” so that it doesn’t scroll out of view for long body content. This may not work well with all browsers.
#     "collapsiblesidebar" : "True", #Add an experimental JavaScript snippet that makes the sidebar collapsible via a button on its side. Doesn’t work with “stickysidebar”.
#     "externalrefs" : "True", #Display external links differently from internal links.
#
#     "footerbgcolor" : white, #Background color for the footer line.
#     "footertextcolor" : dark_grey, #Text color for the footer line.
#
# 	"sidebarbgcolor" : light_grey, #Background color for the sidebar.
#     "sidebarbtncolor" : collapse_button, #Background color for the sidebar collapse button (used when "collapsiblesidebar" is True).
#     "sidebartextcolor" : dark_grey, #Text color for the sidebar.
#     "sidebarlinkcolor" : blue, #Link color for the sidebar.
#
#     "relbarbgcolor" : grey, #Background color for the relation bar.
#     "relbartextcolor" : blue, #Text color for the relation bar.
#     "relbarlinkcolor" : dark_grey, #Link color for the relation bar.
#
#     "bgcolor" : white, #Body background color.
#     "textcolor" : black, #Body text color.
#     "linkcolor" : black, #Body link color.
#     "visitedlinkcolor" : black, #Body color for visited links.
#
#     "headbgcolor" : white, #Background color for headings.
#     "headtextcolor" : black, #Text color for headings.
#     "headlinkcolor" : black,  #Link color for headings.
#
#     "codebgcolor" : light_grey, #Background color for code blocks.
#     "codetextcolor" : black, #Default text color for code blocks, if not set differently by the highlighting style.
#
#     "bodyfont" : "Helvetica, Arial, sans-serif", #Font for normal text.
#     "headfont" : "Helvetica, Arial, sans-serif", #Font for headings.
# }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
# html_sidebars = {
#     '**': [
#         'about.html',
#         'navigation.html',
#         'relations.html',  # needs 'show_related': True theme option to display
#         'searchbox.html',
#         'donate.html',
#     ]
# }


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "vicoconcomdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "virocon.tex", "virocon documentation", "virocon", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "virocon", "virocon documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "virocon",
        "virocon documentation",
        author,
        "virocon",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Napoleon settings
# napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
# napoleon_include_private_with_doc = False
# napoleon_include_special_with_doc = True
# napoleon_use_admonition_for_examples = False
# napoleon_use_admonition_for_notes = False
# napoleon_use_admonition_for_references = False
# napoleon_use_ivar = False
# napoleon_use_param = True
# napoleon_use_rtype = True

# todo settings
# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False
# If this is True, todo emits a warning for each TODO entries.
# todo_emit_warnings = False
# If this is True, todolist produce output without file path and line.
# todo_link_only = False

# Intersphinx options
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.6", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference", None),
    "matplotlib": ("http://matplotlib.org", None),
}



"""
Compute four types of contour with the same dataset.

Namely a direct sampling contour, a highest density contour, an AND contour
and an OR contour. These contours are all defined in the original variable space.

This figure is presented as Figure 2.8 on page 18 in Haselsteiner, A. F. (2022).
Offshore structures under extreme loads: A methodology to determine design loads
[University of Bremen]. https://doi.org/10.26092/elib/1615
"""

import matplotlib.pyplot as plt

from virocon import (
    read_ec_benchmark_dataset,
    AndContour,
    OrContour,
    DirectSamplingContour,
    HighestDensityContour,
    GlobalHierarchicalModel,
    plot_2D_contour,
    get_OMAE2020_V_Hs,
)

# Fit a joint model based on a wind speed - wave height dataset.
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D.txt")
dist_descriptions, fit_descriptions, semantics = get_OMAE2020_V_Hs()
model = GlobalHierarchicalModel(dist_descriptions)
model.fit(data, fit_descriptions)

# Compute the contours.
alpha = 0.1
n = 10000
sample = model.draw_sample(n)
and_contour = AndContour(model, alpha, deg_step=1, sample=sample, allowed_error=0.0001)
or_contour = OrContour(
    model, alpha, deg_step=1, sample=sample, allowed_error=0.0001, lowest_theta=2
)
direct_sampling = DirectSamplingContour(model, alpha, sample=sample, deg_step=10)
highest_density = HighestDensityContour(
    model, alpha, limits=[(0, 20), (0, 20)], deltas=0.02
)

small_sample = model.draw_sample(200)

# Plot the contours.
fig, axs = plt.subplots(2, 2, figsize=[8, 8], sharex=True, sharey=True)
axs[0][0].scatter(
    small_sample[:, 0],
    small_sample[:, 1],
    c="k",
    marker=".",
    alpha=0.3,
    rasterized=True,
)
axs[0][0].plot(
    and_contour.coordinates[0:-1, 0], and_contour.coordinates[0:-1, 1], c="#BB5566"
)
axs[0][1].scatter(
    small_sample[:, 0],
    small_sample[:, 1],
    c="k",
    marker=".",
    alpha=0.3,
    rasterized=True,
)
axs[0][1].plot(
    or_contour.coordinates[0:-3, 0], or_contour.coordinates[0:-3, 1], c="#BB5566"
)
plot_2D_contour(direct_sampling, sample=small_sample, ax=axs[1][0])
plot_2D_contour(highest_density, sample=small_sample, ax=axs[1][1])
titles = [
    ["AND exceedance", "OR exceedance"],
    ["Angular exceedance", "Isodensity exceedance"],
]
for ax2, t2 in zip(axs, titles):
    for ax, t in zip(ax2, t2):
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.set_xticks([])
        ax.set_yticks([])
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.set_title(t)
        ax.set_xlim([0, 25])
        ax.set_ylim([0, 7])
        for line in ax.lines:
            line.set_linewidth(2)
            line.set_color("k")

# plt.savefig("bivariate_exceedance.pdf", dpi=300, bbox_inches='tight', pad_inches=0)
plt.show()



import pytest

import numpy as np

from virocon import (
    WidthOfIntervalSlicer,
    NumberOfIntervalsSlicer,
    PointsPerIntervalSlicer,
)


@pytest.fixture(scope="module")
def test_data():
    """
    Function to create test data.

    Returns
    -------
    Numpy array
        Test data.

    """

    return np.array([1.2, 1.5, 2.4, 2.5, 2.6, 3.1, 3.5, 3.6, 4.0, 5.0])


@pytest.fixture(
    params=[
        (
            {"width": 1, "reference": "left", "min_n_points": 1},
            [1, 2, 3, 4, 5],
            [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6], [4.0], [5.0]],
            [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0], [5.0, 6.0]],
        ),
        (
            {"width": 1, "reference": "left", "right_open": False, "min_n_points": 1},
            [1, 2, 3, 4],
            [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0], [5.0]],
            [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0]],
        ),
        (
            {"width": 1, "min_n_points": 1},
            [1.5, 2.5, 3.5, 4.5, 5.5],
            [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6], [4.0], [5.0]],
            [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0], [5.0, 6.0]],
        ),
        (
            {"width": 1, "right_open": False, "min_n_points": 1},
            [1.5, 2.5, 3.5, 4.5],
            [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0], [5.0]],
            [[1, 2], [2, 3], [3, 4], [4, 5]],
        ),
        (
            {"width": 1, "reference": np.median, "min_n_points": 1},
            [1.35, 2.5, 3.5, 4.0, 5.0],
            [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6], [4.0], [5.0]],
            [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0], [5.0, 6.0]],
        ),
        (
            {"width": 1, "reference": "left", "min_n_points": 2},
            [1, 2, 3],
            [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6]],
            [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0]],
        ),
    ]
)
def woi_params_and_ref(request):
    """
    Fixture that applies the WidthOfIntervalSlicer to the prepared test case.

    Parameters
    ----------
    request : tuple
        Prepared test case for WidthOfIntervalSlicer.

    Returns
    -------
    dict
        Characteristics of interval.

    """

    params = request.param[0]
    references = request.param[1]
    intervals = request.param[2]
    boundaries = request.param[3]
    return {
        "params": params,
        "references": references,
        "intervals": intervals,
        "boundaries": boundaries,
    }


@pytest.fixture(
    params=[
        (
            {"n_intervals": 2, "min_n_points": 1, "min_n_intervals": 2},
            [2.15, 4.05],
            [[1.2, 1.5, 2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0, 5.0]],
            [[1.2, 3.1], [3.1, 5.0]],
        ),
        (
            {
                "n_intervals": 2,
                "include_max": False,
                "min_n_points": 1,
            },
            [2.15, 4.05],
            [[1.2, 1.5, 2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0]],
            [[1.2, 3.1], [3.1, 5.0]],
        ),
        (
            {"n_intervals": 3, "min_n_points": 1},
            [1.8333333333333333, 3.0999999999999996, 4.366666666666666],
            [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6], [4.0, 5.0]],
            [
                [1.2, 2.466666666666667],
                [2.4666666666666663, 3.733333333333333],
                [3.733333333333333, 5.0],
            ],
        ),
        (
            {"n_intervals": 3, "min_n_points": 3, "min_n_intervals": 2},
            [1.8333333333333333, 3.0999999999999996],
            [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6]],
            [[1.2, 2.466666666666667], [2.4666666666666663, 3.733333333333333]],
        ),
        (
            {"n_intervals": 3, "include_max": False, "min_n_points": 1},
            [1.8333333333333333, 3.0999999999999996, 4.366666666666666],
            [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6], [4.0]],
            [
                [1.2, 2.466666666666667],
                [2.4666666666666663, 3.733333333333333],
                [3.733333333333333, 5.0],
            ],
        ),
        (
            {
                "n_intervals": 2,
                "value_range": (0, 5),
                "min_n_points": 1,
            },
            [1.25, 3.75],
            [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6, 4.0, 5.0]],
            [[0, 2.5], [2.5, 5.0]],
        ),
        (
            {
                "n_intervals": 2,
                "reference": np.median,
                "min_n_points": 1,
            },
            [2.4, 3.6],
            [[1.2, 1.5, 2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0, 5.0]],
            [[1.2, 3.1], [3.1, 5.0]],
        ),
        (
            {
                "n_intervals": 3,
                "min_n_points": 3,
            },  # should raise RuntimeError (too few intervals)
            None,
            None,
            None,
        ),
    ]
)
def noi_params_and_ref(request):
    """
    Fixture that applies the NumberOfIntervalsSlicer to the prepared test case.

    Parameters
    ----------
    request : tuple
        Prepared test case for NumberOfIntervalsSlicer.

    Returns
    -------
    dict
        Characteristics of interval.

    """

    params = request.param[0]
    references = request.param[1]
    intervals = request.param[2]
    boundaries = request.param[3]
    return {
        "params": params,
        "references": references,
        "intervals": intervals,
        "boundaries": boundaries,
    }


@pytest.fixture(
    params=[
        (
            {"n_points": 2, "min_n_points": 1},
            [1.35, 2.45, 2.85, 3.55, 4.5],
            [[1.2, 1.5], [2.4, 2.5], [2.6, 3.1], [3.5, 3.6], [4.0, 5.0]],
            [[1.2, 1.95], [1.95, 2.55], [2.55, 3.3], [3.3, 3.8], [3.8, 5.0]],
        ),
        (
            {"n_points": 3, "min_n_points": 1},
            [1.2, 2.4, 3.1, 4.0],
            [[1.2], [1.5, 2.4, 2.5], [2.6, 3.1, 3.5], [3.6, 4.0, 5.0]],
            [[1.2, 1.35], [1.35, 2.55], [2.55, 3.55], [3.55, 5.0]],
        ),
        (
            {"n_points": 3, "last_full": False, "min_n_points": 1},
            [1.5, 2.6, 3.6, 5.0],
            [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1], [3.5, 3.6, 4.0], [5.0]],
            [[1.2, 2.45], [2.45, 3.3], [3.3, 4.5], [4.5, 5.0]],
        ),
        (
            {
                "n_points": 3,
            },
            [2.4, 3.1, 4.0],
            [[1.5, 2.4, 2.5], [2.6, 3.1, 3.5], [3.6, 4.0, 5.0]],
            [[1.5, 2.55], [2.55, 3.55], [3.55, 5.0]],
        ),
        (
            {"n_points": 3, "reference": np.mean, "min_n_points": 1},
            [1.2, 2.1333333333333333, 3.0666666666666664, 4.2],
            [[1.2], [1.5, 2.4, 2.5], [2.6, 3.1, 3.5], [3.6, 4.0, 5.0]],
            [[1.2, 1.35], [1.35, 2.55], [2.55, 3.55], [3.55, 5.0]],
        ),
    ]
)
def ppi_params_and_ref(request):
    """
    Fixture that applies the PointsPerIntervalSlicer to the prepared test case.

    Parameters
    ----------
    request : tuple
        Prepared test case for PointsPerIntervalSlicer.

    Returns
    -------
    dict
        Characteristics of interval.

    """

    params = request.param[0]
    references = request.param[1]
    intervals = request.param[2]
    boundaries = request.param[3]
    return {
        "params": params,
        "references": references,
        "intervals": intervals,
        "boundaries": boundaries,
    }


def test_width_of_interval_slicer(test_data, woi_params_and_ref):
    """
    Function to compare characterisics of reference interval generated by
    WidthOfIntervalSlicer and manually generated interval.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.
    woi_params_and_ref : tuple
        Reference interval generated by the WidthOfIntervalSlicer funciton.

    Returns
    -------
    None.

    """

    params = woi_params_and_ref["params"]
    ref_references = woi_params_and_ref["references"]
    ref_intervals = woi_params_and_ref["intervals"]
    ref_boundaries = woi_params_and_ref["boundaries"]
    slicer = WidthOfIntervalSlicer(**params)

    my_slices, my_references, my_boundaries = slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_intervals_slicer(test_data, noi_params_and_ref):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalsSlicer and manually generated interval.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.
    noi_params_and_ref : tuple
        Reference interval generated by the NumberOfIntervalsSlicer funciton.

    Returns
    -------
    None.

    """

    params = noi_params_and_ref["params"]
    ref_references = noi_params_and_ref["references"]
    ref_intervals = noi_params_and_ref["intervals"]
    ref_boundaries = noi_params_and_ref["boundaries"]
    slicer = NumberOfIntervalsSlicer(**params)

    if ref_references is None:
        with pytest.raises(RuntimeError):
            my_slices, my_references, my_boundaries = slicer.slice_(test_data)
    else:
        my_slices, my_references, my_boundaries = slicer.slice_(test_data)
        my_intervals = [test_data[slice_] for slice_ in my_slices]
        np.testing.assert_almost_equal(my_references, ref_references)
        assert len(my_intervals) == len(ref_intervals)
        for i in range(len(my_intervals)):
            np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
        for i in range(len(my_boundaries)):
            np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_points_per_interval_slicer(test_data, ppi_params_and_ref):
    """
    Function to compare characterisics of reference interval generated by
    PointsPerIntervalSlicer and manually generated interval.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.
    ppi_params_and_ref : tuple
        Reference interval generated by the PointsPerIntervalSlicer funciton.

    Returns
    -------
    None.

    """

    params = ppi_params_and_ref["params"]
    ref_references = ppi_params_and_ref["references"]
    ref_intervals = ppi_params_and_ref["intervals"]
    ref_boundaries = ppi_params_and_ref["boundaries"]
    slicer = PointsPerIntervalSlicer(**params)

    my_slices, my_references, my_boundaries = slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_points_per_interval_slicer_npoints2(test_data):
    """
    Function to compare characterisics of reference interval generated by
    PointsPerIntervalSlicer and manually generated interval.  Specific test
    case: the number of intervals = 2.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    # n_points= 2
    ref_intervals = [[1.2, 1.5], [2.4, 2.5], [2.6, 3.1], [3.5, 3.6], [4.0, 5.0]]
    ref_references = [np.median(inter) for inter in ref_intervals]
    ref_boundaries = [[1.2, 1.95], [1.95, 2.55], [2.55, 3.3], [3.3, 3.8], [3.8, 5.0]]
    n_points_slicer = PointsPerIntervalSlicer(2, min_n_points=1)
    my_slices, my_references, my_boundaries = n_points_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_points_per_interval_slicer_npoints3_lastfull(test_data):
    """
    Function to compare characterisics of reference interval generated by
    PointsPerIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 3, the upper boundary of the last
    interval is inclusive.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    # n_points= 3, last_full=True
    ref_intervals = [[1.2], [1.5, 2.4, 2.5], [2.6, 3.1, 3.5], [3.6, 4.0, 5.0]]
    ref_references = [np.median(inter) for inter in ref_intervals]
    ref_boundaries = [[1.2, 1.35], [1.35, 2.55], [2.55, 3.55], [3.55, 5.0]]
    n_points_slicer = PointsPerIntervalSlicer(3, last_full=True, min_n_points=1)
    my_slices, my_references, my_boundaries = n_points_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_points_per_interval_slicer_npoints3(test_data):
    """
    Function to compare characterisics of reference interval generated by
    PointsPerIntervalSlicer and manually generated interval.  Specific test
    case: the number of intervals = 3, the upper boundary of the last
    interval is not inclusive.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1], [3.5, 3.6, 4.0], [5.0]]
    ref_references = [np.median(inter) for inter in ref_intervals]
    ref_boundaries = [[1.2, 2.45], [2.45, 3.3], [3.3, 4.5], [4.5, 5.0]]
    n_points_slicer = PointsPerIntervalSlicer(3, last_full=False, min_n_points=1)
    my_slices, my_references, my_boundaries = n_points_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_points_per_interval_slicer_npoints3_lastfull_minpoints3(test_data):
    """
    Function to compare characterisics of reference interval generated by
    PointsPerIntervalSlicer and manually generated interval.  Specific test
    case: the number of intervals = 3, the upper boundary of the last
    interval is inclusive and the minimum points per interval = 3.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.5, 2.4, 2.5], [2.6, 3.1, 3.5], [3.6, 4.0, 5.0]]
    ref_references = [np.median(inter) for inter in ref_intervals]
    ref_boundaries = [[1.5, 2.55], [2.55, 3.55], [3.55, 5.0]]
    n_points_slicer = PointsPerIntervalSlicer(3, last_full=True, min_n_points=3)
    my_slices, my_references, my_boundaries = n_points_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_points_per_interval_slicer_npoints3_lastfull_reference_mean(test_data):
    """
    Function to compare characterisics of reference interval generated by
    PointsPerIntervalSlicer and manually generated interval.  Specific test
    case: the number of intervals = 3, the upper boundary of the last
    interval is inclusive and the interval reference is the mean.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2], [1.5, 2.4, 2.5], [2.6, 3.1, 3.5], [3.6, 4.0, 5.0]]
    ref_references = [np.mean(inter) for inter in ref_intervals]
    ref_boundaries = [[1.2, 1.35], [1.35, 2.55], [2.55, 3.55], [3.55, 5.0]]
    n_points_slicer = PointsPerIntervalSlicer(3, reference=np.mean, min_n_points=1)
    my_slices, my_references, my_boundaries = n_points_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints2_include_max(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 2, the upper boundary of the last
    interval is inclusive.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0, 5.0]]
    ref_references = [2.15, 4.05]
    ref_boundaries = [[1.2, 3.1], [3.1, 5.0]]
    number_slicer = NumberOfIntervalsSlicer(2, min_n_points=1)
    my_slices, my_references, my_boundaries = number_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints2(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 2, the upper boundary of the last
    interval is not inclusive.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0]]
    ref_references = [2.15, 4.05]
    ref_boundaries = [[1.2, 3.1], [3.1, 5.0]]
    number_slicer = NumberOfIntervalsSlicer(2, include_max=False, min_n_points=1)
    my_slices, my_references, my_boundaries = number_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints3_include_max(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 3 and the upper boundary of the last
    interval is inclusive.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6], [4.0, 5.0]]
    ref_width = (5 - 1.2) / 3
    ref_references = [
        1.2 + ref_width / 2,
        1.2 + 3 * ref_width / 2,
        1.2 + 5 * ref_width / 2,
    ]
    ref_boundaries = [(c - ref_width / 2, c + ref_width / 2) for c in ref_references]
    number_slicer = NumberOfIntervalsSlicer(3, min_n_points=1)
    my_slices, my_references, my_boundaries = number_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_almost_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints3_numberintervals2_include_max(
    test_data,
):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 3, the upper boundary of the last
    interval is inclusive, the minimum points per intervals = 3 and the
    minimum number of intervals = 2.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6]]
    ref_width = (5 - 1.2) / 3
    ref_references = [1.2 + ref_width / 2, 1.2 + 3 * ref_width / 2]
    ref_boundaries = [(c - ref_width / 2, c + ref_width / 2) for c in ref_references]
    number_slicer = NumberOfIntervalsSlicer(3, min_n_points=3, min_n_intervals=2)
    my_slices, my_references, my_boundaries = number_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_almost_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints3(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 3, the upper boundary of the last
    interval is not inclusive.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6], [4.0]]
    ref_width = (5 - 1.2) / 3
    ref_references = [
        1.2 + ref_width / 2,
        1.2 + 3 * ref_width / 2,
        1.2 + 5 * ref_width / 2,
    ]
    ref_boundaries = [(c - ref_width / 2, c + ref_width / 2) for c in ref_references]
    number_slicer = NumberOfIntervalsSlicer(3, include_max=False, min_n_points=1)
    my_slices, my_references, my_boundaries = number_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_almost_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints2_include_max_valuerange(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 2, the upper boundary of the last
    interval is inclusive, and the the value range used for creating the
    intervals is 0-5.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4], [2.5, 2.6, 3.1, 3.5, 3.6, 4.0, 5.0]]
    ref_references = [1.25, 3.75]
    ref_boundaries = [[0, 2.5], [2.5, 5.0]]
    number_slicer = NumberOfIntervalsSlicer(2, value_range=(0, 5), min_n_points=1)
    my_slices, my_references, my_boundaries = number_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints2_include_max_reference_median(
    test_data,
):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 2, the upper boundary of the last
    interval is inclusive and the reference of the interval is the median.
    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5, 2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0, 5.0]]
    ref_references = [2.4, 3.6]
    ref_boundaries = [[1.2, 3.1], [3.1, 5.0]]
    number_slicer = NumberOfIntervalsSlicer(2, reference=np.median, min_n_points=1)
    my_slices, my_references, my_boundaries = number_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_almost_equal(my_boundaries[i], ref_boundaries[i])


def test_number_of_interval_slicer_numberpoints3_numberintervals3(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the number of intervals = 3, the upper boundary of the last
    interval is inclusive, the minimum points per intervals = 3 and the
    minimum number of intervals = 3.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    number_slicer = NumberOfIntervalsSlicer(3, min_n_points=3, min_n_intervals=3)
    with pytest.raises(
        RuntimeError
    ):  # we expect a RuntimeError if there are less than min_n_intervals intervals
        number_slicer.slice_(test_data)


def test_width_of_interval_slicer_reference_left_right_open(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the reference of the interval is the lower boundary (left) and the
    upper interval boundary is open (right open).

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_references = [1, 2, 3, 4, 5]
    ref_intervals = [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6], [4.0], [5.0]]
    ref_boundaries = [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0], [5.0, 6.0]]
    width_slicer = WidthOfIntervalSlicer(width=1, reference="left", min_n_points=1)
    my_slices, my_references, my_boundaries = width_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_width_of_interval_slicer_reference_left_left_open(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the reference of the interval is the lower boundary (left) and the
    lower interval boundary is open (left open).

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_references = [1, 2, 3, 4]
    ref_intervals = [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0], [5.0]]
    ref_boundaries = [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0]]
    width_slicer = WidthOfIntervalSlicer(
        width=1, reference="left", right_open=False, min_n_points=1
    )
    my_slices, my_references, my_boundaries = width_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_width_of_interval_slicer_reference_center_right_open(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the reference of the interval is the center and the upper interval
    boundary is open (right open).

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_references = [1.5, 2.5, 3.5, 4.5, 5.5]
    ref_intervals = [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6], [4.0], [5.0]]
    ref_boundaries = [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0], [5.0, 6.0]]
    width_slicer = WidthOfIntervalSlicer(width=1, min_n_points=1)
    my_slices, my_references, my_boundaries = width_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_width_of_interval_slicer_reference_center_left_open(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval.  Specific test
    case: the reference of the interval is the center and the lower interval
    boundary is open (left open).

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_references = [1.5, 2.5, 3.5, 4.5]
    ref_intervals = [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6, 4.0], [5.0]]
    ref_boundaries = [[1, 2], [2, 3], [3, 4], [4, 5]]
    width_slicer = WidthOfIntervalSlicer(width=1, right_open=False, min_n_points=1)
    my_slices, my_references, my_boundaries = width_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_width_of_interval_slicer_reference_median_right_open(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the reference of the interval is the median and the upper interval
    boundary is open (right open).

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """

    ref_intervals = [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6], [4.0], [5.0]]
    ref_references = [np.median(x) for x in ref_intervals]
    ref_boundaries = [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0], [4.0, 5.0], [5.0, 6.0]]
    width_slicer = WidthOfIntervalSlicer(width=1, reference=np.median, min_n_points=1)
    my_slices, my_references, my_boundaries = width_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])


def test_width_of_interval_slicer_reference_left_right_open_minpoints(test_data):
    """
    Function to compare characterisics of reference interval generated by
    NumberOfIntervalSlicer and manually generated interval. Specific test
    case: the reference of the interval is the lower boundary (left) and the
    minimum points per interval = 2.

    Parameters
    ----------
    test_data : numpy array
        Data to test the interval slicer.

    Returns
    -------
    None.

    """
    # reference="left", right_open, min_n_points=2
    ref_references = [1, 2, 3]
    ref_intervals = [[1.2, 1.5], [2.4, 2.5, 2.6], [3.1, 3.5, 3.6]]
    ref_boundaries = [[1.0, 2.0], [2.0, 3.0], [3.0, 4.0]]
    width_slicer = WidthOfIntervalSlicer(width=1, reference="left", min_n_points=2)
    my_slices, my_references, my_boundaries = width_slicer.slice_(test_data)
    my_intervals = [test_data[slice_] for slice_ in my_slices]
    np.testing.assert_array_equal(my_references, ref_references)
    assert len(my_intervals) == len(ref_intervals)
    for i in range(len(my_intervals)):
        np.testing.assert_array_equal(my_intervals[i], ref_intervals[i])
    for i in range(len(my_boundaries)):
        np.testing.assert_array_equal(my_boundaries[i], ref_boundaries[i])



"""
Brief example that computes a sea state contour and compares MLE vs WLSQ fitting.
"""

import matplotlib.pyplot as plt
from virocon import (
    read_ec_benchmark_dataset,
    get_OMAE2020_Hs_Tz,
    GlobalHierarchicalModel,
    IFORMContour,
    plot_2D_contour,
)

# Load sea state measurements.
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_A.txt")

# Define the structure of the joint distribution model.
dist_descriptions, fit_descriptions, semantics = get_OMAE2020_Hs_Tz()
model = GlobalHierarchicalModel(dist_descriptions)

# Estimate the model's parameter values with the default method (MLE).
model.fit(data)

# Compute an IFORM contour with a return period of 50 years.
tr = 50  # Return period in years.
ts = 1  # Sea state duration in hours.
alpha = 1 / (tr * 365.25 * 24 / ts)
contour1 = IFORMContour(model, alpha)

# Estimate the model's parameter values using weighted lesat squares.
fit_description_hs = {"method": "wlsq", "weights": "quadratic"}
my_fit_descriptions = [fit_description_hs, None]
model2 = GlobalHierarchicalModel(dist_descriptions)
model2.fit(data, fit_descriptions=my_fit_descriptions)

# Compute an IFORM contour with a return period of 50 years.
tr = 50  # Return period in years.
ts = 1  # Sea state duration in hours.
alpha = 1 / (tr * 365.25 * 24 / ts)
contour2 = IFORMContour(model2, alpha)

# Plot the contours.
fig, axs = plt.subplots(1, 2, figsize=[7.5, 4], sharex=True, sharey=True)
plot_2D_contour(contour1, data, semantics=semantics, swap_axis=True, ax=axs[0])
plot_2D_contour(contour2, data, semantics=semantics, swap_axis=True, ax=axs[1])
titles = ["Maximum likelihood estimation", "Weighted least squares"]
for i, (ax, title) in enumerate(zip(axs, titles)):
    ax.set_title(title)
plt.show()



"""
Distributions for single variables.
"""

import math
import copy

import numpy as np
import scipy.stats as sts

from abc import ABC, abstractmethod
from scipy.optimize import fmin

__all__ = [
    "WeibullDistribution",
    "LogNormalDistribution",
    "NormalDistribution",
    "ExponentiatedWeibullDistribution",
    "GeneralizedGammaDistribution",
    "VonMisesDistribution",
    "ScipyDistribution",
]

# The distributions parameters need to have an order, this order is defined by
# the parameters dict. As of Python 3.7 dicts officially keep their order of creation.
# So this version is a requirement.
# (Though the dict order might work as well in 3.6)


class ConditionalDistribution:
    """
    A conditional probability distribution.

    The conditional distribution uses a Distribution as template and
    dynamically alters its parameters to model the dependence. The
    ConditionalDistribution wraps another distribution. When a method of
    the ConditionalDistribution is called it first computes the distributions
    parameters at given and then calls the corresponding method of the
    distribution with these parameters. Usually the parameters are defined by
    dependence functions of the form dep_func(given) -> param_val.

    Parameters
    ----------
    distribution : Distribution
        The distribution used as template. Its parameters can be replaced with
        dependence functions to model the dependency.
    parameters: float
       A dictionary describing the parameters of distribution. The keys are
       the parameter names, the values are the dependence functions. Every
       parameter that is not fixed in distribution has to be set here.

    Attributes
    ----------
    distribution_class : type
        The class of the distribution used.
    param_names : list-like
        Names of the parameters of the distribution.
    conditional_parameters : dict
        Dictionary of dependence functions for conditional parameters. Parameter names as keys.
    fixed_parameters : dict
        Values of the fixed parameters. The fixed parameters do not change,
        even when fitting them. Parameters as keys.
    distributions_per_interval : list
        Instances of distribution fitted to intervals
    parameters_per_interval : list of dict
        Values of the parameters of the distribution function. Parameter names as keys.
    data_intervals : list of array
        The data that was used to fit the distribution. Split into intervals.
    conditioning_values : array_like
        Realizations of the conditioning variable that where used for fitting.
    conditioning_interval_boundaries : list of tuple
        Boundaries of the intervals the data of the conditioning variable
        was split into.
    """

    def __init__(self, distribution, parameters):
        # allow setting fitting initials on class creation?
        self.distribution = distribution
        self.distribution_class = distribution.__class__
        self.param_names = list(distribution.parameters)
        self.conditional_parameters = {}
        self.fixed_parameters = {}
        self.conditioning_values = None
        # TODO check that dependency functions are not duplicates

        # Check if the parameters dict contains keys/parameters that
        # are not known to the distribution.
        # (use set operations for that purpose)
        unknown_params = set(parameters).difference(self.param_names)
        if len(unknown_params) > 0:
            raise ValueError(
                "Unknown param(s) in parameters."
                f"Known params are {self.param_names}, "
                f"but found {unknown_params}."
            )

        for par_name in self.param_names:
            # is the parameter defined as a dependence function?
            if par_name not in parameters:
                # if it is not a dependence function it must be fixed
                if getattr(distribution, f"f_{par_name}") is None:
                    raise ValueError(
                        "Parameters of the distribution must be "
                        "either defined by a dependence function "
                        f"or fixed, but {par_name} was not defined."
                    )
                else:
                    self.fixed_parameters[par_name] = getattr(
                        distribution, f"f_{par_name}"
                    )
            else:
                # if it is a dependence function it must not be fixed
                if getattr(distribution, f"f_{par_name}") is not None:
                    raise ValueError(
                        "Parameters can be defined by a "
                        "dependence function or by being fixed. "
                        f"But for parameter {par_name} both where given."
                    )
                else:
                    self.conditional_parameters[par_name] = parameters[par_name]

    def __repr__(self):
        dist = "Conditional" + self.distribution_class.__name__
        fixed_params = ", ".join(
            [
                f"f_{par_name}={par_value}"
                for par_name, par_value in self.fixed_parameters.items()
            ]
        )
        cond_params = ", ".join(
            [
                f"{par_name}={repr(par_value)}"
                for par_name, par_value in self.conditional_parameters.items()
            ]
        )
        combined_params = fixed_params + ", " + cond_params
        combined_params = combined_params.strip(", ")
        return f"{dist}({combined_params})"

    def _get_param_values(self, given):
        param_values = {}
        for par_name in self.param_names:
            if par_name in self.conditional_parameters.keys():
                param_values[par_name] = self.conditional_parameters[par_name](given)
            else:
                param_values[par_name] = self.fixed_parameters[par_name]

        return param_values

    def pdf(self, x, given):
        """
        Probability density function for the described random variable.

        With x, value(s) from the sample space of this random variable and
        given value(s) from the sample space of the conditioning random
        variable, pdf(x, given) returns the probability density function at x
        conditioned on given.

        Parameters
        ----------
        x : array_like
            Points at which the pdf is evaluated.
            Shape: 1- dimensional.

        given : float or array_like
           The conditioning value of the conditioning variable i.e. the
           y in x|y.
           Shape: 1-dimensional. Same size as x.

        Returns
        -------
        ndarray
            Probability densities at x conditioned on given.
            Shape: 1- dimensional. Same size as x.


        """

        return self.distribution.pdf(x, **self._get_param_values(given))

    def cdf(self, x, given):
        """
        Cumulative distribution function for the described random variable.

        With x, a realization of this random variable and given a realisation
        of the conditioning random variable, cdf(x, given) returns the
        cumulative distribution function at x conditioned on given.

        Parameters
        ----------
        x : array_like
            Points at which the cdf is evaluated.
            Shape: 1-dimensional.

        given : float or array_like
           The conditioning value of the conditioning variable i.e. the
           y in x|y.
           Shape: 1-dimensional. Same size as x.

        Returns
        -------
        ndarray
            Cumulative distribution function evaluated at x.
            Shape: 1-dimensional. Same size as x.

        """

        return self.distribution.cdf(x, **self._get_param_values(given))

    def icdf(self, prob, given):
        """
        Inverse cumulative distribution function.

        Calculate the inverse cumulative distribution function. Also known as quantile or
        percent-point function. With x, a realization of this random variable
        and given a realisation of the conditioning random variable,
        icdf(x, given) returns the inverse cumulative distribution function at
        x conditioned on given.


        Parameters
        ----------
        prob :
            Probabilities for which the i_cdf is evaluated.
            Shape: 1-dimensional

        given : float or array_like
           The conditioning value of the conditioning variable i.e. the
           y in x|y.
           Shape: 1-dimensional. Same size as prob.

        Returns
        -------
        ndarray or float
            Inverse cumulative distribution function evaluated at given
            probabilities conditioned on given.
            Shape: 1-dimensional. Same size as prob.

        """

        return self.distribution.icdf(prob, **self._get_param_values(given))

    def draw_sample(self, n, given, *, random_state=None):
        """
        Draw a random sample of size n, conditioned on given.


        Parameters
        ----------
        n : int
            Number of observations that shall be drawn.

        given : float or array_like
           The conditioning value of the conditioning variable i.e. the
           y in x|y.
           Shape: TODO
        random_state : {None, int, numpy.random.Generator}, optional
            Can be used to draw a reproducible sample.
        Returns
        -------
        ndarray or float
            Sample of the requested size conditioned on given.

        """

        return self.distribution.draw_sample(
            n, **self._get_param_values(given), random_state=random_state
        )

    def fit(
        self,
        data,
        conditioning_values,
        conditioning_interval_boundaries,
        method=None,
        weights=None,
    ):
        """
        Fit statistical distribution to data.

        Method of estimating the parameters of a probability distribution to
        given data.

        Parameters
        ----------
        data : list of array
            The data that should be used to fit the distribution.
            Realizations of the distributions variable split into intervals.
            One array for each interval containing the data in that interval.
        conditioning_values : array_like
            Realizations of the conditioning variable i.e. the y in x|y.
            One value for each interval in data.
        conditioning_interval_boundaries : list of tuple
            Boundaries of the intervals the data of the conditioning variable
            was split into.
            One 2-element tuple for each interval in data.
        method : str, optional
            The method used to fit the distributions (self.distribution) for each interval.
            Defaults to the distributions default.
        weights :
            The weights used to fit the distributions (self.distribution) for each interval,
            when method is 'wlsq' = weighted least squares.

        """

        self.distributions_per_interval = []
        self.parameters_per_interval = []
        self.data_intervals = data
        self.conditioning_values = np.array(conditioning_values)
        self.conditioning_interval_boundaries = conditioning_interval_boundaries
        # Fit distribution to each interval.
        for interval_data in data:
            # dist = self.distribution_class()
            dist = copy.deepcopy(self.distribution)
            dist.fit(interval_data, method, weights)
            self.distributions_per_interval.append(dist)
            self.parameters_per_interval.append(dist.parameters)

        # Fit dependence functions.
        fitted_dependence_functions = {}
        for par_name, dep_func in self.conditional_parameters.items():
            x = self.conditioning_values
            y = [params[par_name] for params in self.parameters_per_interval]
            dep_func.fit(x, y)
            fitted_dependence_functions[par_name] = dep_func

        self.conditional_parameters = fitted_dependence_functions


class Distribution(ABC):
    """
    Abstract base class for distributions.

    Models the probabilities of occurrence for different possible
    (environmental) events.

    """

    def __repr__(self):
        dist_name = self.__class__.__name__
        param_names = self.parameters.keys()
        set_params = {}
        for par_name in param_names:
            f_attr = getattr(self, f"f_{par_name}")
            if f_attr is not None:
                set_params[f"f_{par_name}"] = f_attr
            else:
                set_params[par_name] = getattr(self, par_name)

        params = ", ".join(
            [f"{par_name}={par_value}" for par_name, par_value in set_params.items()]
        )

        return f"{dist_name}({params})"

    @property
    @abstractmethod
    def parameters(self):
        """
        Parameters of the probability distribution.

        Dict of the form: {"<parameter_name>" : <parameter_value>, ...}

        """

        return {}

    @abstractmethod
    def cdf(self, x, *args, **kwargs):
        """
        Cumulative distribution function.

        """

    @abstractmethod
    def pdf(self, x, *args, **kwargs):
        """
        Probability density function.

        """

    @abstractmethod
    def icdf(self, prob, *args, **kwargs):
        """
        Inverse cumulative distribution function.

        """

    @abstractmethod
    def draw_sample(self, n, *args, random_state=None, **kwargs):
        """
        Draw a random sample of length n.

        n : float
            Number of observations that shall be drawn.
        args, kwargs : float, optional
            Parameter of the distribution
        random_state : {None, int, numpy.random.Generator}, optional
            Can be used to draw a reproducible sample.
        """

    def fit(self, data, method="mle", weights=None):
        """
        Fit the distribution to the sampled data.

        data : array_like
            The observed data to fit the distribution.
        method : str, optional
            The method used for fitting.

             - "mle" = maximum-likelihood estimation (default)
             - "mom" = method of moments
             - "lsq" = least squares method
             - "wlsq" = weighted least squares

        weights : None, str, array_like,
            The weights to use for weighted least squares fitting. Ignored otherwise.
            Defaults to None = equal weights.
            Can be either an array_like with one weight for each point in data or a str.
            Valid options for str are: 'linear', 'quadratic', 'cubic'.
        """

        if method.lower() == "mle":
            self._fit_mle(data)
        elif method.lower() == "mom":
            self._fit_mom(data)
        elif method.lower() == "lsq" or method.lower() == "wlsq":
            self._fit_lsq(data, weights)
        else:
            raise ValueError(
                f"Unknown fit method '{method}'. "
                "Only maximum likelihood estimation (keyword: mle) ",
                "method of moments (keywork: mom)",
                "and (weighted) least squares (keyword: (w)lsq) are supported.",
            )

    @abstractmethod
    def _fit_mle(self, data):
        """Fit the distribution using maximum likelihood estimation."""

    def _fit_mom(self, data):
        """Fit the distribution using the method of moments."""
        raise NotImplementedError(
            f"Fitting by method of moments is not implemented for class {self.__class__.__name__}."
        )

    @abstractmethod
    def _fit_lsq(self, data, weights):
        """Fit the distribution using (weighted) least squares."""

    @staticmethod
    def _get_rvs_size(n, pars):
        # Returns the size parameter for the scipy rvs method.
        # If there are any iterable pars it is a tuple,
        # otherwise n is returned.
        at_least_one_iterable = False
        par_length = 0
        for par in pars:
            try:
                _ = iter(par)
                at_least_one_iterable = True
                par_length = len(par)
            except TypeError:
                pass

        if at_least_one_iterable:
            return (n, par_length)
        else:
            return n


class WeibullDistribution(Distribution):
    """
    A weibull distribution.

    The distributions probability density function is given by [1]_ :

    :math:`f(x) = \\frac{\\beta}{\\alpha} \\left (\\frac{x-\\gamma}{\\alpha} \\right)^{\\beta -1} \\exp \\left[-\\left( \\frac{x-\\gamma}{\\alpha} \\right)^{\\beta} \\right]`

    Parameters
    ----------
    alpha : float
        Scale parameter of the weibull distribution. Defaults to 1.
    beta : float
        Shape parameter of the weibull distribution. Defaults to 1.
    gamma : float
        Location parameter of the weibull distribution (3-parameter weibull
        distribution). Defaults to 0.
    f_alpha : float
        Fixed scale parameter of the weibull distribution (e.g. given physical
        parameter). If this parameter is set, alpha is ignored. The fixed
        parameter does not change, even when fitting the distribution. Defaults to None.
    f_beta : float
       Fixed shape parameter of the weibull distribution (e.g. given physical
       parameter). If this parameter is set, beta is ignored. The fixed parameter
       does not change, even when fitting the distribution. Defaults to None.
    f_gamma : float
        Fixed location parameter of the weibull distribution (e.g. given physical
        parameter). If this parameter is set, gamma is ignored. The fixed
        parameter does not change, even when fitting the distribution. Defaults to None.

    References
    ----------
    .. [1] Haselsteiner, A.F.; Ohlendorf, J.H.; Wosniok, W.; Thoben, K.D.(2017)
        Deriving environmental contours from highest density regions.
        Coastal Engineering 123 (2017) 42–51.

    """

    def __init__(
        self, alpha=1, beta=1, gamma=0, f_alpha=None, f_beta=None, f_gamma=None
    ):
        self.alpha = alpha if f_alpha is None else f_alpha  # scale
        self.beta = beta if f_beta is None else f_beta  # shape
        self.gamma = gamma if f_gamma is None else f_gamma  # loc
        self.f_alpha = f_alpha
        self.f_beta = f_beta
        self.f_gamma = f_gamma

    @property
    def parameters(self):
        return {"alpha": self.alpha, "beta": self.beta, "gamma": self.gamma}

    def _get_scipy_parameters(self, alpha, beta, gamma):
        if alpha is None:
            alpha = self.alpha
        if beta is None:
            beta = self.beta
        if gamma is None:
            gamma = self.gamma
        return beta, gamma, alpha  # shape, loc, scale

    def cdf(self, x, alpha=None, beta=None, gamma=None):
        """
        Cumulative distribution function.

        Parameters
        ----------
        x : array_like,
            Points at which the cdf is evaluated.
            Shape: 1-dimensional.
        alpha : float, optional
            The scale parameter. Defaults to self.alpha.
        beta : float, optional
            The shape parameter. Defaults to self.beta.
        gamma: float, optional
            The location parameter . Defaults to self.gamma.

        """

        scipy_par = self._get_scipy_parameters(alpha, beta, gamma)
        return sts.weibull_min.cdf(x, *scipy_par)

    def icdf(self, prob, alpha=None, beta=None, gamma=None):
        """
        Inverse cumulative distribution function.

        Parameters
        ----------
        prob : array_like
            Probabilities for which the i_cdf is evaluated.
            Shape: 1-dimensional
        alpha : float, optional
            The scale parameter. Defaults to self.aplha .
        beta : float, optional
            The shape parameter. Defaults to self.beta.
        gamma: float, optional
            The location parameter . Defaults to self.gamma.

        """

        scipy_par = self._get_scipy_parameters(alpha, beta, gamma)
        return sts.weibull_min.ppf(prob, *scipy_par)

    def pdf(self, x, alpha=None, beta=None, gamma=None):
        """
        Probability density function.

        Parameters
        ----------
        x : array_like,
            Points at which the pdf is evaluated.
            Shape: 1-dimensional.
        alpha_ : float, optional
            The scale parameter. Defaults to self.alpha.
        beta : float, optional
            The shape parameter. Defaults to self.beta.
        gamma: float, optional
            The location parameter . Defaults to self.gamma.

        """

        scipy_par = self._get_scipy_parameters(alpha, beta, gamma)
        return sts.weibull_min.pdf(x, *scipy_par)

    def draw_sample(self, n, alpha=None, beta=None, gamma=None, *, random_state=None):
        scipy_par = self._get_scipy_parameters(alpha, beta, gamma)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return sts.weibull_min.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mle(self, sample):
        self._fit_scipy(sample, method="MLE")

    def _fit_mom(self, sample):
        self._fit_scipy(sample, method="MM")

    def _fit_scipy(self, sample, method="MLE"):
        p0 = {"alpha": self.alpha, "beta": self.beta, "gamma": self.gamma}

        fparams = {"method": method}
        if self.f_beta is not None:
            fparams["f0"] = self.f_beta
        if self.f_gamma is not None:
            fparams["floc"] = self.f_gamma
        if self.f_alpha is not None:
            fparams["fscale"] = self.f_alpha

        self.beta, self.gamma, self.alpha = sts.weibull_min.fit(
            sample, p0["beta"], loc=p0["gamma"], scale=p0["alpha"], **fparams
        )

    def _fit_lsq(self, data, weights):
        raise NotImplementedError()


class LogNormalDistribution(Distribution):
    """
    A Lognormal Distribution.

    The distributions probability density function is given by [2]_:

    :math:`f(x) = \\frac{1}{x\\widetilde{\\sigma} \\sqrt{2\\pi}}\\exp \\left[ \\frac{-(\\ln x - \\widetilde{\\mu})^2}{2\\widetilde{\\sigma}^2}\\right]`


    Parameters
    ----------
    mu : float
        Mean parameter of the corresponding normal distribution.
        Defaults to 0.
    sigma : float
        Standard deviation of the corresponding normal distribution.
        Defaults to 1.
    f_mu : float
        Fixed parameter mu of the lognormal distribution (e.g. given physical
        parameter). If this parameter is set, mu is ignored. The fixed
        parameter does not change, even when fitting the distribution. Defaults to None.
    f_sigma : float
       Fixed parameter sigma of the lognormal distribution (e.g. given
       physical parameter). If this parameter is set, sigma is ignored. The
       fixed parameter does not change, even when fitting the distribution. Defaults to None.

    References
    ----------
    .. [2] Forbes, C.; Evans, M.; Hastings, N; Peacock, B. (2011)
        Statistical Distributions, 4th Edition, Published by
        John Wiley & Sons, Inc., Hoboken, New Jersey.,
        Pages 131-132
    """

    def __init__(self, mu=0, sigma=1, f_mu=None, f_sigma=None):
        self.mu = mu if f_mu is None else f_mu
        self.sigma = sigma if f_sigma is None else f_sigma  # shape
        self.f_mu = f_mu
        self.f_sigma = f_sigma
        # self.scale = math.exp(mu)

    @property
    def parameters(self):
        return {"mu": self.mu, "sigma": self.sigma}

    @property
    def _scale(self):
        return np.exp(self.mu)

    @_scale.setter
    def _scale(self, val):
        self.mu = np.log(val)

    def _get_scipy_parameters(self, mu, sigma):
        if mu is None:
            scale = self._scale
        else:
            scale = np.exp(mu)
        if sigma is None:
            sigma = self.sigma
        return sigma, 0, scale  # shape, loc, scale

    def cdf(self, x, mu=None, sigma=None):
        """
        Cumulative distribution function.

        Parameters
        ----------
        x : array_like,
            Points at which the cdf is evaluated.
            Shape: 1-dimensional.
        mu : float, optional
            The variance parameter. Defaults to self.mu .
        sigma : float, optional
            The shape parameter. Defaults to self.sigma .

        """

        scipy_par = self._get_scipy_parameters(mu, sigma)
        return sts.lognorm.cdf(x, *scipy_par)

    def icdf(self, prob, mu=None, sigma=None):
        """
        Inverse cumulative distribution function.

        Parameters
        ----------
        prob : Probabilities for which the i_cdf is evaluated.
            Shape: 1-dimensional
        mu : float, optional
            The variance parameter. Defaults to self.mu .
        sigma : float, optional
            The shape parameter. Defaults to self.sigma .

        """

        scipy_par = self._get_scipy_parameters(mu, sigma)
        return sts.lognorm.ppf(prob, *scipy_par)

    def pdf(self, x, mu=None, sigma=None):
        """
        Probability density function.

        Parameters
        ----------
        x : array_like,
            Points at which the pdf is evaluated.
            Shape: 1-dimensional.
        mu : float, optional
            The variance parameter. Defaults to self.mu .
        sigma : float, optional
            The shape parameter. Defaults to self.sigma .

        """

        scipy_par = self._get_scipy_parameters(mu, sigma)
        return sts.lognorm.pdf(x, *scipy_par)

    def draw_sample(self, n, mu=None, sigma=None, *, random_state=None):
        scipy_par = self._get_scipy_parameters(mu, sigma)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return sts.lognorm.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mle(self, sample):
        self._fit_scipy(sample, method="MLE")

    def _fit_mom(self, sample):
        self._fit_scipy(sample, method="MM")

    def _fit_scipy(self, sample, method="MLE"):
        p0 = {"scale": self._scale, "sigma": self.sigma}

        fparams = {"floc": 0, "method": method}

        if self.f_sigma is not None:
            fparams["f0"] = self.f_sigma
        if self.f_mu is not None:
            fparams["fscale"] = math.exp(self.f_mu)

        # scale0 = math.exp(p0["mu"])
        self.sigma, _, self._scale = sts.lognorm.fit(
            sample, p0["sigma"], scale=p0["scale"], **fparams
        )
        # self.mu = math.log(self._scale)

    def _fit_lsq(self, data, weights):
        raise NotImplementedError()


class NormalDistribution(Distribution):
    """
    A Normal (Gaussian) Distribution.

    The distributions probability density function is given by [3]_:

    :math:`f(x) = \\frac{1}{{\\sigma} \\sqrt{2\\pi}} \\exp \\left( - \\frac{( x - \\mu)^2}{2\\sigma^2}\\right)`


    Parameters
    ----------
    mu : float
        Location parameter, the mean.
        Defaults to 0.
    sigma : float
        Scale parameter, the standard deviation.
        Defaults to 1.
    f_mu : float
        Fixed parameter mu of the normal distribution (e.g. given physical
        parameter). If this parameter is set, mu is ignored. The fixed
        parameter does not change, even when fitting the distribution. Defaults to None.
    f_sigma : float
       Fixed parameter sigma of the normal distribution (e.g. given
       physical parameter). If this parameter is set, sigma is ignored. The fixed
       parameter does not change, even when fitting the distribution. Defaults to None.

    References
    ----------
    .. [3] Forbes, C.; Evans, M.; Hastings, N; Peacock, B. (2011)
        Statistical Distributions, 4th Edition, Published by
        John Wiley & Sons, Inc., Hoboken, New Jersey.,
        Page 143
    """

    def __init__(self, mu=0, sigma=1, f_mu=None, f_sigma=None):
        self.mu = mu if f_mu is None else f_mu  # location
        self.sigma = sigma if f_sigma is None else f_sigma  # scale
        self.f_mu = f_mu
        self.f_sigma = f_sigma

    @property
    def parameters(self):
        return {"mu": self.mu, "sigma": self.sigma}

    def _get_scipy_parameters(self, mu, sigma):
        if mu is None:
            loc = self.mu
        else:
            loc = mu
        if sigma is None:
            scale = self.sigma
        else:
            scale = self.sigma
        return loc, scale  # loc, scale

    def cdf(self, x, mu=None, sigma=None):
        """
        Cumulative distribution function.

        Parameters
        ----------
        x : array_like,
            Points at which the cdf is evaluated.
            Shape: 1-dimensional.
        mu : float, optional
            The location parameter. Defaults to self.mu .
        sigma : float, optional
            The scale parameter. Defaults to self.sigma .

        """
        scipy_par = self._get_scipy_parameters(mu, sigma)
        return sts.norm.cdf(x, *scipy_par)

    def icdf(self, prob, mu=None, sigma=None):
        """
        Inverse cumulative distribution function.

        Parameters
        ----------
        prob : Probabilities for which the i_cdf is evaluated.
            Shape: 1-dimensional
        mu : float, optional
            The location parameter. Defaults to self.mu .
        sigma : float, optional
            The scale parameter. Defaults to self.sigma .

        """
        scipy_par = self._get_scipy_parameters(mu, sigma)
        return sts.norm.ppf(prob, *scipy_par)

    def pdf(self, x, mu=None, sigma=None):
        """
        Probability density function.

        Parameters
        ----------
        x : array_like,
            Points at which the pdf is evaluated.
            Shape: 1-dimensional.
        mu : float, optional
            The location parameter. Defaults to self.mu .
        sigma : float, optional
            The scale parameter. Defaults to self.sigma .

        """
        scipy_par = self._get_scipy_parameters(mu, sigma)
        return sts.norm.pdf(x, *scipy_par)

    def draw_sample(self, n, mu=None, sigma=None, *, random_state=None):
        scipy_par = self._get_scipy_parameters(mu, sigma)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return sts.norm.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mle(self, sample):
        self._fit_scipy(sample, method="MLE")

    def _fit_mom(self, sample):
        self._fit_scipy(sample, method="MM")

    def _fit_scipy(self, sample, method="MLE"):
        p0 = {"loc": self.mu, "scale": self.sigma}

        fparams = {"method": method}

        if self.f_mu is not None:
            fparams["floc"] = self.f_mu
        if self.f_sigma is not None:
            fparams["fscale"] = self.f_sigma

        self.mu, self.sigma = sts.norm.fit(
            sample, loc=p0["loc"], scale=p0["scale"], **fparams
        )

    def _fit_lsq(self, data, weights):
        raise NotImplementedError()


class LogNormalNormFitDistribution(LogNormalDistribution):
    # https://en.wikipedia.org/wiki/Log-normal_distribution#Estimation_of_parameters
    """
    A Lognormal Distribution.
    The distributions probability density function is given by:

    :math:`f(x) = \\frac{1}{x\\widetilde{\\sigma} \\sqrt{2\\pi}}\\exp \\left[ \\frac{-(\\ln x - \\widetilde{\\mu})^2}{2\\widetilde{\\sigma}^2}\\right]`


    Parameters
    ----------
    mu : float
        Mean parameter of the corresponding normal distribution.
        Defaults to 0.
    sigma : float
        Variance parameter of the corresponding normal distribution.
        Defaults to 1.
    f_mu : float
        Fixed parameter mu of the lognormal distribution (e.g. given physical
        parameter). If this parameter is set, mu is ignored. The fixed
        parameter does not change, even when fitting the distribution. Defaults to None.
    f_sigma : float
       Fixed parameter sigma of the lognormal distribution (e.g. given
       physical parameter). If this parameter is set, sigma is ignored. The
       fixed parameter does not change, even when fitting the distribution.
       Defaults to None.

    """

    def __init__(self, mu_norm=0, sigma_norm=1, f_mu_norm=None, f_sigma_norm=None):
        self.mu_norm = mu_norm if f_mu_norm is None else f_mu_norm
        self.sigma_norm = sigma_norm if f_sigma_norm is None else f_sigma_norm
        self.f_mu_norm = f_mu_norm
        self.f_sigma_norm = f_sigma_norm

    @property
    def parameters(self):
        return {"mu_norm": self.mu_norm, "sigma_norm": self.sigma_norm}

    @property
    def mu(self):
        return self.calculate_mu(self.mu_norm, self.sigma_norm)

    @staticmethod
    def calculate_mu(mu_norm, sigma_norm):
        return np.log(mu_norm / np.sqrt(1 + sigma_norm**2 / mu_norm**2))
        # return np.log(mu_norm**2 * np.sqrt(1 / (sigma_norm**2 + mu_norm**2)))

    @property
    def sigma(self):
        return self.calculate_sigma(self.mu_norm, self.sigma_norm)

    @staticmethod
    def calculate_sigma(mu_norm, sigma_norm):
        # return np.sqrt(np.log(1 + sigma_norm**2 / mu_norm**2))
        return np.sqrt(np.log(1 + (sigma_norm**2 / mu_norm**2)))

    def _get_scipy_parameters(self, mu_norm, sigma_norm):
        if (mu_norm is None) != (sigma_norm is None):
            raise RuntimeError(
                "mu_norm and sigma_norm have to be passed both or not at all"
            )

        if mu_norm is None:
            scale = self._scale
            sigma = self.sigma
        else:
            sigma = self.calculate_sigma(mu_norm, sigma_norm)
            mu = self.calculate_mu(mu_norm, sigma_norm)
            scale = np.exp(mu)
        return sigma, 0, scale  # shape, loc, scale

    def cdf(self, x, mu_norm=None, sigma_norm=None):
        scipy_par = self._get_scipy_parameters(mu_norm, sigma_norm)
        return sts.lognorm.cdf(x, *scipy_par)

    def icdf(self, prob, mu_norm=None, sigma_norm=None):
        scipy_par = self._get_scipy_parameters(mu_norm, sigma_norm)
        return sts.lognorm.ppf(prob, *scipy_par)

    def pdf(self, x, mu_norm=None, sigma_norm=None):
        scipy_par = self._get_scipy_parameters(mu_norm, sigma_norm)
        return sts.lognorm.pdf(x, *scipy_par)

    def draw_sample(self, n, mu_norm=None, sigma_norm=None, *, random_state=None):
        scipy_par = self._get_scipy_parameters(mu_norm, sigma_norm)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return sts.lognorm.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mom(self, sample):
        # This is needed to avoid calling the corresponding implementation of the parent class.
        raise NotImplementedError()

    def _fit_mle(self, sample):
        if self.f_mu_norm is None:
            self.mu_norm = np.mean(sample)
        else:
            self.mu_norm = self.f_mu_norm

        if self.f_sigma_norm is None:
            self.sigma_norm = np.std(sample, ddof=1)
        else:
            self.sigma_norm = self.f_sigma_norm

    def _fit_lsq(self, data, weights):
        raise NotImplementedError()


class ExponentiatedWeibullDistribution(Distribution):
    """
    An exponentiated Weibull distribution.

    The parametrization used is the same as described by
    Haselsteiner et al. (2019) [4]_.  The distributions cumulative distribution
    function is given by:

    :math:`F(x) = \\left[ 1- \\exp \\left(-\\left( \\frac{x}{\\alpha} \\right)^{\\beta} \\right) \\right] ^{\\delta}`

    Parameters
    ----------
    alpha : float
        Scale parameter of the exponentiated weibull distribution. Defaults
        to 1.
    beta : float
        First shape parameter of the exponentiated weibull distribution.
        Defaults to 1.
    delta : float
        Second shape parameter of the exponentiated weibull distribution.
        Defaults to 1.
    f_alpha : float
        Fixed alpha parameter of the weibull distribution (e.g. given physical
        parameter). If this parameter is set, alpha is ignored. The fixed
        parameter does not change, even when fitting the distribution. Defaults to None.
    f_beta : float
       Fixed beta parameter of the weibull distribution (e.g. given physical
       parameter). If this parameter is set, beta is ignored. The fixed
       parameter does not change, even when fitting the distribution. Defaults to None.
    f_delta : float
        Fixed delta parameter of the weibull distribution (e.g. given physical
        parameter). If this parameter is set, delta is ignored. The fixed
        parameter does not change, even when fitting the distribution. Defaults to None.

    References
    ----------
    .. [4] Haselsteiner, A.F.; Thoben, K.D. (2019)
        Predicting wave heights for marine design by prioritizing extreme events in
        a global model, Renewable Energy, Volume 156, August 2020,
        Pages 1146-1157; https://doi.org/10.1016/j.renene.2020.04.112

    """

    @property
    def parameters(self):
        return {"alpha": self.alpha, "beta": self.beta, "delta": self.delta}

    def __init__(
        self, alpha=1, beta=1, delta=1, f_alpha=None, f_beta=None, f_delta=None
    ):
        self.alpha = alpha if f_alpha is None else f_alpha  # scale
        self.beta = beta if f_beta is None else f_beta  # shape
        self.delta = delta if f_delta is None else f_delta  # shape2
        self.f_alpha = f_alpha
        self.f_beta = f_beta
        self.f_delta = f_delta
        # In scipy the order of the shape parameters is reversed:
        # a ^= delta
        # c ^= beta

    def _get_scipy_parameters(self, alpha, beta, delta):
        if alpha is None:
            alpha = self.alpha
        if beta is None:
            beta = self.beta
        if delta is None:
            delta = self.delta
        return delta, beta, 0, alpha  # shape1, shape2, loc, scale

    def cdf(self, x, alpha=None, beta=None, delta=None):
        scipy_par = self._get_scipy_parameters(alpha, beta, delta)
        return sts.exponweib.cdf(x, *scipy_par)

    def icdf(self, prob, alpha=None, beta=None, delta=None):
        scipy_par = self._get_scipy_parameters(alpha, beta, delta)
        return sts.exponweib.ppf(prob, *scipy_par)

    def pdf(self, x, alpha=None, beta=None, delta=None):
        # x = np.asarray(x, dtype=float)  # If x elements are int we cannot use np.nan .
        # This changes x which is unepexted behaviour!
        # x[x<=0] = np.nan  # To avoid warnings with negative and 0-values, use NaN.
        # TODO ensure for all pdf that no nan come up?
        x_greater_zero = np.where(x > 0, x, np.nan)
        scipy_par = self._get_scipy_parameters(alpha, beta, delta)
        _pdf = sts.exponweib.pdf(x_greater_zero, *scipy_par)
        if _pdf.shape == ():  # x was scalar
            if np.isnan(_pdf):
                _pdf = 0
        else:
            _pdf[np.isnan(_pdf)] = 0
        return _pdf

    def draw_sample(self, n, alpha=None, beta=None, delta=None, *, random_state=None):
        scipy_par = self._get_scipy_parameters(alpha, beta, delta)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return sts.exponweib.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mle(self, sample):
        self._fit_scipy(sample, method="MLE")

    def _fit_mom(self, sample):
        self._fit_scipy(sample, method="MM")

    def _fit_scipy(self, sample, method="MLE"):
        p0 = {"alpha": self.alpha, "beta": self.beta, "delta": self.delta}

        fparams = {"floc": 0, "method": method}

        if self.f_delta is not None:
            fparams["f0"] = self.f_delta
        if self.f_beta is not None:
            fparams["f1"] = self.f_beta
        if self.f_alpha is not None:
            fparams["fscale"] = self.f_alpha

        self.delta, self.beta, _, self.alpha = sts.exponweib.fit(
            sample, p0["delta"], p0["beta"], scale=p0["alpha"], **fparams
        )

    def _fit_lsq(self, data, weights):
        # Based on Appendix A. in https://arxiv.org/pdf/1911.12835.pdf
        x = np.sort(np.asarray_chkfinite(data))
        if weights is None:
            weights = np.ones_like(x)
        elif isinstance(weights, str):
            if weights.lower() == "linear":
                weights = x / np.sum(x)
            elif weights.lower() == "quadratic":
                weights = x**2 / np.sum(x**2)
            elif weights.lower() == "cubic":
                weights = x**3 / np.sum(x**3)
            else:
                raise ValueError(f"Unsupported value for weights={weights}.")
        else:
            try:
                _ = iter(weights)
                weights = np.asarray_chkfinite(weights)
            except TypeError:
                raise ValueError(f"Unsupported value for weights={weights}.")

        n = len(x)
        p = (np.arange(1, n + 1) - 0.5) / n

        fixed = {}
        if self.f_alpha is not None:
            fixed["alpha"] = self.f_alpha
        if self.f_beta is not None:
            fixed["beta"] = self.f_beta
        if self.f_delta is not None:
            fixed["delta"] = self.f_delta
        if len(fixed) == 0:
            fixed = None

        if fixed is not None:
            if "delta" in fixed and not ("alpha" in fixed or "beta" in fixed):
                self.delta = fixed["delta"]
                self.alpha, self.beta = self._estimate_alpha_beta(
                    self.delta,
                    x,
                    p,
                    weights,
                )
            else:
                raise NotImplementedError()
        else:
            delta0 = self.delta
            self.delta = fmin(
                self._wlsq_error, delta0, disp=False, args=(x, p, weights)
            )[0]

            self.alpha, self.beta = self._estimate_alpha_beta(
                self.delta,
                x,
                p,
                weights,
            )

    @staticmethod
    def _estimate_alpha_beta(delta, x, p, w, falpha=None, fbeta=None):
        # As x = 0 causes problems when x_star is calculated, zero-elements
        # are not considered in the parameter estimation.
        indices = np.nonzero(x)
        x = x[indices]
        p = p[indices]
        w = w[indices]

        # First, transform x and p to get a linear relationship.
        x_star = np.log10(x)
        p_star = np.log10(-np.log(1 - p ** (1 / delta)))

        # Estimate the parameters alpha_hat and beta_hat.
        p_star_bar = np.sum(w * p_star)
        x_star_bar = np.sum(w * x_star)
        b_hat_dividend = np.sum(w * p_star * x_star) - p_star_bar * x_star_bar
        b_hat_divisor = np.sum(w * p_star**2) - p_star_bar**2
        b_hat = b_hat_dividend / b_hat_divisor
        a_hat = x_star_bar - b_hat * p_star_bar
        alpha_hat = 10**a_hat
        beta_hat = b_hat_divisor / b_hat_dividend  # beta_hat = 1 / b_hat

        return alpha_hat, beta_hat

    @staticmethod
    def _wlsq_error(delta, x, p, w, return_alpha_beta=False, falpha=None, fbeta=None):
        # As x = 0 causes problems when x_star is calculated, zero-elements
        # are not considered in the parameter estimation.
        indices = np.nonzero(x)
        x = x[indices]
        p = p[indices]
        w = w[indices]

        alpha_hat, beta_hat = ExponentiatedWeibullDistribution._estimate_alpha_beta(
            delta, x, p, w
        )

        # Compute the weighted least squares error.
        x_hat = alpha_hat * (-np.log(1 - p ** (1 / delta))) ** (1 / beta_hat)
        wlsq_error = np.sum(w * (x - x_hat) ** 2)

        return wlsq_error


class GeneralizedGammaDistribution(Distribution):
    """
    A 3-parameter generalized Gamma distribution.

    The parametrization is orientated on [5]_ :

    :math:`f(x) = \\frac{ \\lambda^{cm} cx^{cm-1} \\exp \\left[ - \\left(\\lambda x^{c} \\right) \\right] }{\\Gamma(m)}`

    Parameters
    ----------
    m : float
        First shape parameter of the generalized Gamma distribution. Defaults to 1.
    c : float
        Second shape parameter of the generalized Gamma distribution. Defaults
        to 1.
    lambda\_ : float
        Scale parameter of the generalized Gamma distribution.
        Defaults to 1.
    f_m : float
        Fixed shape parameter of the generalized Gamma distribution (e.g.
        given physical parameter). If this parameter is set, m is ignored. The
        fixed parameter does not change, even when fitting the distribution.
        Defaults to None.
    f_c : float
       Fixed second shape parameter of the generalized Gamma distribution (e.g.
       given  physical parameter). If this parameter is set, c is ignored. The
       fixed parameter does not change, even when fitting the distribution.
       Defaults to None.
    f_lambda\_ : float
        Fixed reciprocal scale parameter of the generalized Gamma distribution
        (e.g. given physical parameter). If this parameter is set, lambda\_ is
        ignored. The fixed parameter does not change, even when fitting the distribution.
        Defaults to None.

    References
    ----------
    .. [5] M.K. Ochi, New approach for estimating the severest sea state from
        statistical data , Coast. Eng. Chapter 38 (1992)
        pp. 512-525.

    """

    def __init__(self, m=1, c=1, lambda_=1, f_m=None, f_c=None, f_lambda_=None):
        self.m = m if f_m is None else f_m  # shape
        self.c = c if f_c is None else f_c  # shape
        self.lambda_ = lambda_ if f_lambda_ is None else f_lambda_  # reciprocal scale
        self.f_m = f_m
        self.f_c = f_c
        self.f_lambda_ = f_lambda_

    @property
    def parameters(self):
        return {"m": self.m, "c": self.c, "lambda_": self.lambda_}

    @property
    def _scale(self):
        return 1 / (self.lambda_)

    @_scale.setter
    def _scale(self, val):
        self.lambda_ = 1 / val

    def _get_scipy_parameters(self, m, c, lambda_):
        if m is None:
            m = self.m
        if c is None:
            c = self.c
        if lambda_ is None:
            scipy_scale = self._scale
        else:
            scipy_scale = 1 / lambda_
        return m, c, 0, scipy_scale  # shape1, shape2, location=0, reciprocal scale

    def cdf(self, x, m=None, c=None, lambda_=None):
        """
        Cumulative distribution function.

        Parameters
        ----------
        x : array_like,
            Points at which the cdf is evaluated.
            Shape: 1-dimensional.
        m : float, optional
            First shape parameter. Defaults to self.m.
        c : float, optional
            The second shape parameter. Defaults to self.c.
        lambda_: float, optional
            The reciprocal scale parameter . Defaults to self.lambda\_.

        """

        scipy_par = self._get_scipy_parameters(m, c, lambda_)
        return sts.gengamma.cdf(x, *scipy_par)

    def icdf(self, prob, m=None, c=None, lambda_=None):
        """
        Inverse cumulative distribution function.

        Parameters
        ----------
        prob : array_like
            Probabilities for which the i_cdf is evaluated.
            Shape: 1-dimensional
        m : float, optional
            First shape parameter. Defaults to self.m.
        c : float, optional
            The second shape parameter. Defaults to self.c.
        lambda_: float, optional
            The reciprocal scale parameter . Defaults to self.lambda\_.

        """

        scipy_par = self._get_scipy_parameters(m, c, lambda_)
        return sts.gengamma.ppf(prob, *scipy_par)

    def pdf(self, x, m=None, c=None, lambda_=None):
        """
        Probability density function.

        Parameters
        ----------
        x : array_like,
            Points at which the pdf is evaluated.
            Shape: 1-dimensional.
        m : float, optional
            First shape parameter. Defaults to self.m.
        c : float, optional
            The second shape parameter. Defaults to self.k.
        lambda_: float, optional
            The reciprocal scale parameter . Defaults to self.lambda\_.

        """

        scipy_par = self._get_scipy_parameters(m, c, lambda_)
        return sts.gengamma.pdf(x, *scipy_par)

    def draw_sample(self, n, m=None, c=None, lambda_=None, random_state=None):
        scipy_par = self._get_scipy_parameters(m, c, lambda_)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return sts.gengamma.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mle(self, sample):
        self._fit_scipy(sample, method="MLE")

    def _fit_mom(self, sample):
        self._fit_scipy(sample, method="MM")

    def _fit_scipy(self, sample, method="MLE"):
        p0 = {"m": self.m, "c": self.c, "scale": self._scale}

        fparams = {"floc": 0, "method": method}

        if self.f_m is not None:
            fparams["fshape1"] = self.f_m
        if self.f_c is not None:
            fparams["fshape2"] = self.f_c
        if self.f_lambda_ is not None:
            fparams["fscale"] = 1 / (self.f_lambda_)

        self.m, self.c, _, self._scale = sts.gengamma.fit(
            sample, p0["m"], p0["c"], scale=p0["scale"], **fparams
        )

    def _fit_lsq(self, data, weights):
        raise NotImplementedError()


class VonMisesDistribution(Distribution):
    """
    A Von Mises (Circular Norm) Distribution.

    The distributions probability density function is given by [1]_:

    :math:`f(x) = \\frac{\\exp{(\\kappa \\cos{(x - \\mu)})}}{2 \\pi I_0(\\kappa)}`

    The distribution is used to model wind-wave misalignment in [2]_.
    Being a circular norm distribution it can be used to model direction.

    Parameters
    ----------
    kappa: float
        Shape parameter
        Defaults to 1.
    mu : float
        Location parameter, the mean.
        Defaults to 0.
    f_kappa : float
       Fixed parameter kappa of the von mises distribution (e.g. given
       physical parameter). If this parameter is set, kappa is ignored.
       Defaults to None.
    f_mu : float
        Fixed parameter mu of the von mises distribution (e.g. given physical
        parameter). If this parameter is set, mu is ignored. Defaults to
        None.


    References
    ----------
    .. [1] Mardia, Kantilal; Jupp, Peter E. (1999).
       Directional Statistics. Wiley. ISBN 978-0-471-95333-3.

    .. [2] Stewart G M, Robertson A, Jonkman J and Lackner M A 2016
       The creation of a comprehensive metocean data set for offshore
       wind turbine simulations: Comprehensive metocean data set
       Wind Energy 19 1151–9
    """

    def __init__(self, kappa=1, mu=0, f_kappa=None, f_mu=None):
        self.kappa = kappa  # shpae
        self.mu = mu  # location
        self.f_kappa = f_kappa
        self.f_mu = f_mu

    @property
    def parameters(self):
        return {"kappa": self.kappa, "mu": self.mu}

    def _get_scipy_parameters(self, kappa, mu):
        if mu is None:
            loc = self.mu
        else:
            loc = mu
        if kappa is None:
            shape = self.kappa
        else:
            shape = kappa
        return shape, loc

    def cdf(self, x, kappa=None, mu=None):
        """
        Cumulative distribution function.

        Parameters
        ----------
        x : array_like,
            Points at which the cdf is evaluated.
            Shape: 1-dimensional.
        kappa : float, optional
            The shape parameter. Defaults to self.kappa .
        mu : float, optional
            The location parameter. Defaults to self.mu .

        """
        scipy_par = self._get_scipy_parameters(kappa, mu)
        return sts.vonmises.cdf(x, *scipy_par)

    def icdf(self, prob, kappa=None, mu=None):
        """
        Inverse cumulative distribution function.

        Parameters
        ----------
        prob : Probabilities for which the i_cdf is evaluated.
            Shape: 1-dimensional
        kappa : float, optional
            The shape parameter. Defaults to self.kappa .
        mu : float, optional
            The location parameter. Defaults to self.mu .

        """
        scipy_par = self._get_scipy_parameters(kappa, mu)
        return sts.vonmises.ppf(prob, *scipy_par)

    def pdf(self, x, kappa=None, mu=None):
        """
        Probability density function.

        Parameters
        ----------
        x : array_like,
            Points at which the pdf is evaluated.
            Shape: 1-dimensional.
        kappa : float. optional,
            The shape parameter. Defaults to self.kappa .
        mu : float, optional
            The location parameter. Defaults to self.mu .

        """
        scipy_par = self._get_scipy_parameters(kappa, mu)
        return sts.vonmises.pdf(x, *scipy_par)

    def draw_sample(self, n, kappa=None, mu=None, random_state=None):
        scipy_par = self._get_scipy_parameters(kappa, mu)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return sts.vonmises.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mle(self, sample):
        self._fit_scipy(sample, method="MLE")

    def _fit_mom(self, sample):
        self._fit_scipy(sample, method="MM")

    def _fit_scipy(self, sample, method="MLE"):
        p0 = {"shape": self.kappa, "loc": self.mu}

        fparams = {"fscale": 1, "method": method}

        if self.f_mu is not None:
            fparams["floc"] = self.f_mu
        if self.f_kappa is not None:
            fparams["fshape"] = self.f_kappa

        (
            self.kappa,
            self.mu,
            _,
        ) = sts.vonmises.fit(sample, p0["shape"], loc=p0["loc"], scale=1, **fparams)

    def _fit_lsq(self, data, weights):
        raise NotImplementedError()


class ScipyDistribution(Distribution):
    """
    A base class used for virocon distributions derived from a scipy distribution.

    To create a new virocon distribution inherit from ScipyDistribution
    and overwrite either `scipy_dist` or `scipy_dist_name`.

    Attributes
    ----------
    scipy_dist : sts.rv_continuous
        A continuous distribution defined in scipy.
        (Or acting like a distribution defined there.)
    scipy_dist_name : str
        The name of a continuous distribution defined in scipy.
    """

    scipy_dist: sts.rv_continuous
    scipy_dist_name: str

    def __init__(self, *args, **kwargs):
        try:
            self.scipy_dist = getattr(sts, self.scipy_dist_name)
        except AttributeError:
            pass
        try:
            self.scipy_dist
        except AttributeError:
            raise TypeError(
                f"Can't instantiate abstract class {self.__class__.__name__}. "
                "Overwrite either scipy_dist or scipy_dist_name in subclass."
            )
        self.scipy_dist_name = self.scipy_dist.name
        self._param_names = self._list_scipy_parameters(self.scipy_dist)
        self._set_default_parameter_values()

        # read parameter from args: (shape(s), loc, scale)
        assert len(args) <= len(self._param_names)
        for arg, par_name in zip(args, self._param_names):
            setattr(self, par_name, arg)

        assert len(kwargs) <= 2 * len(self._param_names)
        for key, arg in kwargs.items():
            if key in self._param_names:
                setattr(self, key, arg)
            elif key.startswith("f_") and key[2:] in self._param_names:
                setattr(self, key, arg)
                setattr(self, key[2:], arg)
            else:
                raise TypeError(
                    f"{self.__class__.__name__} got an unexpected keyword argument '{key}'"
                )

    #  https://stackoverflow.com/a/53640468
    @staticmethod
    def _list_scipy_parameters(distribution):
        """List parameters for scipy.stats.distribution.
        # Arguments
            distribution: a string or scipy.stats distribution object.
        # Returns
            A list of distribution parameter strings.
        """
        if isinstance(distribution, str):
            distribution = getattr(sts, distribution)
        if distribution.shapes:
            parameters = [name.strip() for name in distribution.shapes.split(",")]
        else:
            parameters = []
        parameters += ["loc", "scale"]
        return parameters

    def _set_default_parameter_values(self):
        for par_name in self._param_names:
            if par_name == "loc":
                setattr(self, par_name, 0)
            else:
                setattr(self, par_name, 1)
            setattr(self, f"f_{par_name}", None)

    @property
    def parameters(self):
        return {par_name: getattr(self, par_name) for par_name in self._param_names}

    def _get_scipy_parameters(self, *args, **kwargs):
        args_with_default = list(self.parameters.values())
        for i, arg in enumerate(args):
            if arg is not None:
                args_with_default[i] = arg

        for key, arg in kwargs.items():
            try:
                idx = self._param_names.index(key)
            except ValueError as e:
                raise ValueError("Unknown parameter name") from e
            args_with_default[idx] = arg

        return args_with_default

    def cdf(self, x, *args, **kwargs):
        scipy_par = self._get_scipy_parameters(*args, **kwargs)
        return self.scipy_dist.cdf(x, *scipy_par)

    def icdf(self, prob, *args, **kwargs):
        scipy_par = self._get_scipy_parameters(*args, **kwargs)
        return self.scipy_dist.ppf(prob, *scipy_par)

    def pdf(self, x, *args, **kwargs):
        scipy_par = self._get_scipy_parameters(*args, **kwargs)
        return self.scipy_dist.pdf(x, *scipy_par)

    def draw_sample(self, n, *args, random_state=None, **kwargs):
        scipy_par = self._get_scipy_parameters(*args, **kwargs)
        rvs_size = self._get_rvs_size(n, scipy_par)
        return self.scipy_dist.rvs(*scipy_par, size=rvs_size, random_state=random_state)

    def _fit_mle(self, sample):
        self._fit_scipy(sample, method="MLE")

    def _fit_mom(self, sample):
        self._fit_scipy(sample, method="MM")

    def _fit_scipy(self, sample, method="MLE"):
        # Split initial parameter values into positional shape parameters and loc and scale.
        p0 = [v for k, v in self.parameters.items() if k != "loc" and k != "scale"]
        loc0 = self.parameters.get("loc", 0)
        scale0 = self.parameters.get("scale", 1)

        fparams = {"method": method}
        for par_name in self._param_names:
            val = getattr(self, f"f_{par_name}")
            if val is not None:
                fparams[f"f{par_name}"] = val

        if len(fparams) == len(self.parameters):
            return  # nothing to do

        params = self.scipy_dist.fit(sample, *p0, loc=loc0, scale=scale0, **fparams)

        assert len(params) == len(self._param_names)
        for par_name, par_value in zip(self._param_names, params):
            setattr(self, par_name, par_value)

    def _fit_lsq(self, data, weights):
        raise NotImplementedError()



"""
Use a wind speed - wave height dataset, fit the join distribution
proposed by Haselsteiner et al. (2020) and compare four different
contour construction methods.

A 50-year wind speed - wave height contour is for example used in
offshore wind turbine design.

Haselsteiner, A. F., Sander, A., Ohlendorf, J.-H., & Thoben,
K.-D. (2020). Global hierarchical models for wind and wave contours:
Physical interpretations of the dependence functions. Proc. 39th
International Conference on Ocean, Offshore and Arctic Engineering
(OMAE 2020). https://doi.org/10.1115/OMAE2020-18668
"""

import matplotlib.pyplot as plt

from virocon import (
    read_ec_benchmark_dataset,
    GlobalHierarchicalModel,
    get_OMAE2020_V_Hs,
    IFORMContour,
    ISORMContour,
    DirectSamplingContour,
    HighestDensityContour,
    plot_2D_contour,
)

# Load a wind speed - significant wave height dataset.
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D.txt")

# Define the structure of the joint distribution model.
dist_descriptions, fit_descriptions, semantics = get_OMAE2020_V_Hs()
model = GlobalHierarchicalModel(dist_descriptions)

# Fit the model to the data (estimate the model's parameter values).
model.fit(data, fit_descriptions)

# Compute four types of contours with a return period of 50 years.
state_duration = 1  # hours
return_period = 50  # years
alpha = state_duration / (return_period * 365.25 * 24)
iform = IFORMContour(model, alpha)
isorm = ISORMContour(model, alpha)
direct_sampling = DirectSamplingContour(model, alpha)
highest_density = HighestDensityContour(model, alpha)

# Plot the contours on top of the metocean data.
fig, axs = plt.subplots(4, 1, figsize=[4, 12], sharex=True, sharey=True)
plot_2D_contour(iform, sample=data, semantics=semantics, ax=axs[0])
plot_2D_contour(isorm, sample=data, semantics=semantics, ax=axs[1])
plot_2D_contour(direct_sampling, sample=data, semantics=semantics, ax=axs[2])
plot_2D_contour(highest_density, sample=data, semantics=semantics, ax=axs[3])
titles = ["IFORM", "ISORM", "Direct sampling", "Highest density"]
for i, (ax, title) in enumerate(zip(axs, titles)):
    ax.set_title(title)
    if i < 3:
        ax.set_xlabel("")

plt.tight_layout()
plt.show()



"""
Models for the joint probability distribution.
"""

import warnings

from abc import ABC, abstractmethod
from typing import Callable

import numpy as np

import scipy.integrate as integrate

from virocon.distributions import ConditionalDistribution
from virocon.intervals import NumberOfIntervalsSlicer

__all__ = ["GlobalHierarchicalModel", "TransformedModel"]


class MaxIterationWarning(RuntimeWarning):
    """The maximum number of iterations was reached."""


class CouldNotSampleError(RuntimeError):
    """Could not draw sample for the supplied parameters."""


class MultivariateModel(ABC):
    """
    Abstract base class for MultivariateModel.

    Statistical model of multiple variables.

    """

    @abstractmethod
    def pdf(self, *args, **kwargs):
        """
        Probability density function.

        """
        pass

    def cdf(self, x):
        """
        Cumulative distribution function.


        Parameters
        ----------
        x : array_like
            Points at which the cdf is evaluated.
            Shape: (n, n_dim), where n is the number of points at which the
            cdf should be evaluated.
        """

        x = np.atleast_2d(np.asarray_chkfinite(x))

        n_dim = self.n_dim
        integral_order = list(range(n_dim))

        def get_integral_func():
            arg_order = integral_order

            def integral_func(*args):
                assert len(args) == n_dim
                # sort arguments as expected by pdf (the models order)
                x = np.array(args)[np.argsort(arg_order)].reshape((1, n_dim))
                return self.pdf(x)

            return integral_func

        lower_integration_limits = [0] * n_dim

        integral_func = get_integral_func()

        p = np.empty(len(x))
        for i in range(len(x)):
            integration_limits = [
                (lower_integration_limits[j], x[i, j]) for j in range(n_dim)
            ]

            p[i], error = integrate.nquad(integral_func, integration_limits)

        return p

    @abstractmethod
    def marginal_pdf(self, *args, **kwargs):
        """
        Marginal probability density function.

        """
        pass

    def marginal_icdf(self, p, dim, precision_factor=1):
        """
        Marginal inverse cumulative distribution function.

        Estimates the marginal icdf by drawing a Monte-Carlo sample.
        Parameters
        ----------
        p : array_like
            Probabilities for which the icdf is evaluated.
            Shape: 1-dimensional
        dim : int
            The dimension for which the marginal is calculated.
        precision_factor : float
            Precision factor that determines the size of the sample to draw.
            A sample is drawn of which on average precision_factor * 100
            realizations exceed the quantile. Minimum sample size is 100000.
            Defaults to 1.0
        """
        pass

        p = np.array(p)

        p_min = np.min(p)
        p_max = np.max(p)
        nr_exceeding_points = 100 * precision_factor
        p_small = np.min([p_min, 1 - p_max])
        n = int((1 / p_small) * nr_exceeding_points)
        n = max([n, 100000])
        sample = self.draw_sample(n)
        x = np.quantile(sample[:, dim], p)
        return x

    @abstractmethod
    def draw_sample(self, *args, random_state=None, **kwargs):
        """
        Draw a random sample of length n.

        """
        pass

    def conditional_cdf(self, x, dim, given, *, random_state=None):
        # assert len(x) == len(given)
        # TODO optimize: reuse sample for equal givens
        p = np.empty_like(x)
        n = 100_000
        for i, (x_val, given_val) in enumerate(zip(x, given)):
            try:
                sample = self.conditional_sample(
                    n, dim, given_val, random_state=random_state
                )
                p[i] = (sample <= x_val).sum() / n
            except CouldNotSampleError:
                p[i] = 0

        return p

    def conditional_icdf(
        self, p, dim, given, precision_factor=1.0, *, random_state=None
    ):
        # assert len(p) == len(given)
        # TODO optimize: reuse sample for equal givens
        x = np.empty_like(p)
        # n = 100_000
        nr_exceeding_points = 100 * precision_factor
        for i, (p_val, given_val) in enumerate(zip(p, given)):
            # p_small = np.min([p_min, 1 - p_max])
            p_small = p_val if p_val < 0.5 else 1 - p_val
            n = (1 / p_small) * nr_exceeding_points
            n = min([max([n, 100_000]), 10_000_000])
            n = int(n)
            try:
                sample = self.conditional_sample(
                    n, dim, given_val, random_state=random_state
                )
                x[i] = np.quantile(sample, p_val)
            except CouldNotSampleError:
                x[i] = 0  # TODO

        return x

    def conditional_sample(
        self, n, dim, given, *, random_state=None, max_iter=100, debug=False
    ):
        # rejection sampling
        # https://github.com/peteroupc/peteroupc.github.io/blob/master/randomfunc.md#rejection-sampling-with-a-pdf-like-function
        # https://github.com/peteroupc/peteroupc.github.io/blob/master/randomfunc.md#Rejection_Sampling
        # https://gist.github.com/rsnemmen/d1c4322d2bc3d6e36be8

        # TODO use given properly
        def get_pdf_like(dim, given):
            """
            let y = given
            We want f(x| y).
            => f(x| y) = f(x, y) / f(y)
            f(y) is const as y is const.
            => f(y) is just a normalization constant
            which is not necessary for rejection sampling
            so we need to use dim and given to get f(x, y)
            """

            def pdf_like(x):
                nonlocal dim, given, self
                n_dim = self.n_dim
                given = np.atleast_1d(given)
                x_hat = np.empty((len(x), n_dim))
                j = 0
                for i in range(n_dim):
                    if i == dim:
                        x_hat[:, i] = x
                    else:
                        x_hat[:, i] = given[j]
                        j += 1
                pdf = self.pdf(x_hat)
                return pdf

            return pdf_like

        pdf = get_pdf_like(dim, given)

        # if random_state already is a np.random.Generator,
        # default_rng returns it unaltered
        rng = np.random.default_rng(random_state)

        # TODO is there a better way to set the minimum value than to set it to close to zero?
        # We use 1e-16 to avoid divide by zero in transform in pdf of transformed model
        x_min = 1e-16

        # TODO is there a better way to find an upper limit of the distribution?
        # For the variables wind speed, significant wave height, peak period and steepness
        # we usually have density approaching 0 at values between 0.07 (steepness) and 50 (wind speed in m/s)
        #
        # Current implementation: Iteratively find a reasonable x_max above which density is close to zero.
        highest_possible_x_max = 100
        lowest_possible_x_max = 0.05
        x_max = highest_possible_x_max
        f_threshold = (
            1e-7  # 10-7 is losely based on Figure 3.4 in DOI: 10.26092/elib/1615
        )
        multiply_xmax_per_iteration = 0.7
        while pdf([x_max]) < f_threshold:
            if x_max * multiply_xmax_per_iteration > lowest_possible_x_max:
                x_max = multiply_xmax_per_iteration * x_max
            else:
                warnings.warn(
                    f"Using the smallest possible x_max value of {lowest_possible_x_max} in conditional_sample although the density is still lower than {f_threshold}"
                )
                x_max = lowest_possible_x_max
                break

        # find max value of pdf
        x = np.linspace(x_min, x_max, 1000)
        y = pdf(x)
        f_min = 0.0
        f_max = y.max() * 1.001  # Use a limit greater than the maximum of the data

        n_counter = 0
        reject_counter = 0
        partial_samples = []

        for i in range(max_iter):
            if n_counter >= n:
                break
            tmp_n = max([(n - n_counter) * 10, n])
            x = rng.uniform(x_min, x_max, size=tmp_n)
            y = rng.uniform(f_min, f_max, size=tmp_n)

            accept_mask = y < pdf(x)

            n_accept = accept_mask.sum()
            n_reject = tmp_n - n_accept

            n_counter += n_accept
            reject_counter += n_reject

            if n_accept > 0:
                partial_samples.append(x[accept_mask])

        if debug:
            print(f"acceptance rate: {n_counter / (n_counter + reject_counter)}")

        if i == max_iter - 1:
            warnings.warn(
                f"Max iterations was reached, sample size is only {n_counter}. Acceptance rate was {n_counter / (n_counter + reject_counter)} and x_max was {x_max}.",
                MaxIterationWarning,
            )
            if len(partial_samples) == 0:
                raise CouldNotSampleError(
                    "Could not draw sample for the supplied parameters."
                )
            return np.concatenate(partial_samples)
        else:
            return np.concatenate(partial_samples)[:n]


class GlobalHierarchicalModel(MultivariateModel):
    """
    Hierarchical probabilistic model.

    Probabilistic model that covers the complete range of an environmental
    variable ("global"), following a particular hierarchical dependence
    structure. The factorization describes a hierarchy where a random
    variable with index i can only depend upon random variables with
    indices less than i [1]_ .

    Parameters
    ----------
    dist_descriptions : dict
        Description of the distributions.

    Attributes
    ----------
    distributions : list
        The distributions used in the GlobalHierachicalModel.
    conditional_on : list
        Indicates the dependencies between the variables of the model. One
        entry per distribution/dimension. Contains either None or int. If the
        ith entry is None, the ith distribution is unconditional. If the ith
        entry is an int j, the ith distribution depends on the jth dimension.
    interval_slicers : list
        One interval slicer per dimension. The interval slicer used for
        slicing the intervals of the corresponding dimension, when necessary
        during fitting.
    n_dim : int
        The number of dimensions, i.e. the number of variables of the model.


    References
    ----------
    .. [1] Haselsteiner, A.F.; Sander, A.; Ohlendorf, J.H.; Thoben, K.D. (2020)
        Global hierarchical models for wind and wave contours: physical
        interpretations of the dependence functions. OMAE 2020, Fort Lauderdale,
        USA. Proceedings of the 39th International Conference on Ocean,
        Offshore and Arctic Engineering.

    Examples
    --------
    Create a Hs-Tz model and fit it to the available data. The following
    example follows the methodology of OMAE2020 [1]_ .

    Example 1.1:

    Load the predefined OMAE 2020 model of Hs-Tz.

    >>> from virocon import (GlobalHierarchicalModel, get_OMAE2020_Hs_Tz,
    ...                      read_ec_benchmark_dataset)
    >>> data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D_1year.txt")
    >>> dist_descriptions, fit_descriptions, semantics = get_OMAE2020_Hs_Tz()
    >>> ghm = GlobalHierarchicalModel(dist_descriptions)
    >>> ghm.fit(data, fit_descriptions=fit_descriptions)

    Example 1.2:

    Create the same OMEA 2020 model manually.

    >>> from virocon import (DependenceFunction, ExponentiatedWeibullDistribution,
    ...                      LogNormalDistribution, WidthOfIntervalSlicer)

    >>> def _asymdecrease3(x, a, b, c):
    ...     return a + b / (1 + c * x)

    >>> def _lnsquare2(x, a, b, c):
    ...     return np.log(a + b * np.sqrt(np.divide(x, 9.81)))

    >>> bounds = [(0, None),
    ...           (0, None),
    ...           (None, None)]

    >>> sigma_dep = DependenceFunction(_asymdecrease3, bounds=bounds)
    >>> mu_dep = DependenceFunction(_lnsquare2, bounds=bounds)

    >>> dist_description_hs = {"distribution" : ExponentiatedWeibullDistribution(),
    ...                        "intervals" : WidthOfIntervalSlicer(width=0.5,
    ...                                                            min_n_points=50)
    ...                       }

    >>> dist_description_tz = {"distribution" : LogNormalDistribution(),
    ...                        "conditional_on" : 0,
    ...                        "parameters" : {"sigma" : sigma_dep,
    ...                                        "mu": mu_dep,
    ...                                        },
    ...                       }


    >>> dist_descriptions = [dist_description_hs, dist_description_tz]

    >>> fit_description_hs = {"method" : "wlsq", "weights" : "quadratic"}
    >>> fit_descriptions = [fit_description_hs, None]

    >>> semantics = {"names" : ["Significant wave height", "Zero-crossing wave period"],
    ...              "symbols" : ["H_s", "T_z"],
    ...              "units" : ["m", "s"]
    ...              }

    >>> ghm = GlobalHierarchicalModel(dist_descriptions)
    >>> ghm.fit(data, fit_descriptions=fit_descriptions)

    """

    _dist_description_keys = {
        "distribution",
        "intervals",
        "conditional_on",
        "parameters",
    }

    def __init__(self, dist_descriptions):
        self.distributions = []
        self.conditional_on = []
        self.interval_slicers = []
        self.n_dim = len(dist_descriptions)
        self._check_dist_descriptions(dist_descriptions)
        for dist_desc in dist_descriptions:
            # dist_class = dist_desc["distribution"]
            dist = dist_desc["distribution"]
            self.interval_slicers.append(
                dist_desc.get("intervals", NumberOfIntervalsSlicer(n_intervals=10))
            )

            if "conditional_on" in dist_desc:
                self.conditional_on.append(dist_desc["conditional_on"])
                dist = ConditionalDistribution(dist, dist_desc["parameters"])
                self.distributions.append(dist)
            else:
                self.conditional_on.append(None)
                self.distributions.append(dist)

        if self.conditional_on[0] is not None:
            raise RuntimeError(
                "Illegal state encountered. The first dimension "
                "has to be independent, but was conditional on "
                f"{self.conditional_on[0]}."
            )

    def __repr__(self):
        name = "GlobalHierarchicalModel"
        dists = repr(self.distributions)
        # dists = dists.replace("), ", "),\n")
        cond_on = repr(self.conditional_on)

        return f"{name}(distributions={dists}, conditional_on={cond_on})"

    def _check_dist_descriptions(self, dist_descriptions):
        for i, dist_desc in enumerate(dist_descriptions):
            if "distribution" not in dist_desc:
                raise ValueError(
                    "Mandatory key 'distribution' missing in "
                    f"dist_description for dimension {i}"
                )

            if "conditional_on" in dist_desc and "parameters" not in dist_desc:
                raise ValueError(
                    "For conditional distributions the "
                    "dist_description key 'parameters' "
                    f"is mandatory but was missing for dimension {i}."
                )

            unknown_keys = set(dist_desc).difference(self._dist_description_keys)
            if len(unknown_keys) > 0:
                raise ValueError(
                    "Unknown key(s) in dist_description for "
                    f"dimension {i}."
                    f"Known keys are {self._dist_description_keys}, "
                    f"but found {unknown_keys}."
                )

    def _split_in_intervals(self, data, dist_idx, conditioning_idx):
        slicer = self.interval_slicers[conditioning_idx]
        conditioning_data = data[:, conditioning_idx]
        interval_slices, interval_centers, interval_boundaries = slicer.slice_(
            conditioning_data
        )

        dist_data = [data[int_slice, dist_idx] for int_slice in interval_slices]

        return dist_data, interval_centers, interval_boundaries

    def _check_and_fill_fit_desc(self, fit_descriptions):
        default_fit_desc = {"method": "mle", "weights": None}
        if fit_descriptions is None:
            fit_descriptions = [default_fit_desc for i in range(self.n_dim)]

        else:
            if len(fit_descriptions) != self.n_dim:
                raise ValueError(
                    "fit_description must have one entry per dimension, but "
                    f"a length of {len(fit_descriptions)} != {self.n_dim} was found."
                )

            for i in range(len(fit_descriptions)):
                if fit_descriptions[i] is None:
                    fit_descriptions[i] = default_fit_desc
                else:
                    if "method" not in fit_descriptions[i]:
                        raise ValueError(
                            "Mandatory key 'method' missing in "
                            f"fit_description for dimension {i}."
                        )
                    if "weights" not in fit_descriptions[i]:
                        fit_descriptions[i]["weights"] = None

        return fit_descriptions

    def fit(self, data, fit_descriptions=None):
        """
        Fit joint model to data.

        Method of estimating the parameters of a probability distribution to
        given data.

        Parameters
        ----------
        data : array-like
            The data that should be used to fit the joint model.
            Shape: (number of realizations, n_dim)
        fit_description : dict
            Description of the fit method. Defaults to None.

        """

        data = np.array(data)

        fit_descriptions = self._check_and_fill_fit_desc(fit_descriptions)

        if data.shape[-1] != self.n_dim:
            raise ValueError(
                "The dimension of data does not match the "
                "dimension of the model. "
                f"The model has {self.n_dim} dimensions, "
                f"but the data has {data.shape[-1]} dimensions."
            )

        for i in range(self.n_dim):
            dist = self.distributions[i]
            conditioning_idx = self.conditional_on[i]
            fit_method = fit_descriptions[i]["method"]
            weights = fit_descriptions[i]["weights"]

            if conditioning_idx is None:
                dist.fit(data[:, i], fit_method, weights)
            else:
                (
                    dist_data,
                    conditioning_data,
                    conditioning_interval_boundaries,
                ) = self._split_in_intervals(data, i, conditioning_idx)
                # dist data  is a list of ndarray
                # and conditioning_data is a list of interval points
                dist.fit(
                    dist_data,
                    conditioning_data,
                    conditioning_interval_boundaries,
                    fit_method,
                    weights,
                )

            self.distributions[i] = (
                dist  # TODO is the writeback necessary? -> probably not
            )

    def pdf(self, x):
        """
        Probability density function.

        Parameters
        ----------
        x : array_like
            Points at which the pdf is evaluated.
            Shape: (n, n_dim), where n is the number of points at which the
            pdf should be evaluated.

        """

        # Ensure that x is a 2D numpy array.
        x = np.array(x)
        if x.ndim == 1:
            x = np.array([x])

        x = np.asarray_chkfinite(x)
        fs = np.empty_like(x)

        fs[:, 0] = self.distributions[0].pdf(x[:, 0])

        for i in range(1, self.n_dim):
            if self.conditional_on[i] is None:
                fs[:, i] = self.distributions[i].pdf(x[:, i])
            else:
                cond_idx = self.conditional_on[i]
                fs[:, i] = self.distributions[i].pdf(x[:, i], given=x[:, cond_idx])

        return np.prod(fs, axis=-1)

    def marginal_pdf(self, x, dim):
        """
        Marginal probability density function.

        Parameters
        ----------
        x : array_like
            Points at which the pdf is evaluated.
            Shape: 1-dimensional
        dim : int
            The dimension for which the marginal is calculated.

        """

        # x = x.reshape((-1, 1))
        if self.conditional_on[dim] is None:
            # the distribution is not conditional -> it is the marginal
            return self.distributions[dim].pdf(x)

        # the distribution is conditional
        # thus we integrate over the joint pdf to get the marginal

        # TODO check size of x

        n_dim = self.n_dim
        integral_order = list(range(n_dim))
        del integral_order[dim]  # we do not integrate over the dim'th variable
        integral_order = integral_order[::-1]  # we integrate over last dimensions first

        # scipy.integrate.nquad expects one argument per dimension
        # thus we have to wrap the (joint) pdf
        def get_integral_func():
            arg_order = integral_order + [dim]

            def integral_func(*args):
                assert len(args) == n_dim
                # sort arguments as expected by pdf (the models order)
                # arguments = list(args)[:-1]
                # arguments.append(args[-1][0])
                x = np.array(args)[np.argsort(arg_order)].reshape((1, n_dim))
                return self.pdf(x)

            return integral_func

        # TODO make limits a property of the distributions?
        # "for var in integral_order append limits"
        # but for now we simplify that all vars have the same limits
        limit = (0, np.inf)
        limits = [limit] * (n_dim - 1)

        f = np.empty_like(x)
        integral_func = get_integral_func()
        for i, x_i in enumerate(x):
            result, _ = integrate.nquad(integral_func, ranges=limits, args=[x_i])
            f[i] = result
        return f

    def marginal_cdf(self, x, dim):
        """
        Marginal cumulative distribution function.

        Parameters
        ----------
        x : array_like
            Points at which the cdf is evaluated.
            Shape: 1-dimensional
        dim : int
            The dimension for which the marginal is calculated.

        """

        # x = x.reshape((-1, 1))
        if self.conditional_on[dim] is None:
            # the distribution is not conditional -> it is the marginal
            return self.distributions[dim].cdf(x)

        # the distribution is conditional
        # thus we integrate over the joint pdf to get the marginal pdf
        # and then integrate the marginal pdf to get the marginal cdf

        # TODO check size of x

        n_dim = self.n_dim
        integral_order = list(range(n_dim))
        del integral_order[dim]
        integral_order = integral_order[::-1]  # we integrate over last dimensions first
        integral_order = integral_order + [
            dim
        ]  # finally we integrate over the dim'th var

        # scipy.integrate.nquad expects one argument per dimension
        # thus we have to wrap the (joint) pdf
        def get_integral_func():
            arg_order = integral_order

            def integral_func(*args):
                assert len(args) == n_dim
                # sort arguments as expected by pdf (the models order)
                # arguments = list(args)[:-1]
                # arguments.append(args[-1][0])
                x = np.array(args)[np.argsort(arg_order)].reshape((1, n_dim))
                return self.pdf(x)

            return integral_func

        # TODO make limits (or lower limit) a property of the distributions?
        limit = (0, np.inf)
        limits = [limit] * (n_dim - 1)

        F = np.empty_like(x)
        integral_func = get_integral_func()
        for i, x_i in enumerate(x):
            result, _ = integrate.nquad(integral_func, ranges=limits + [(0, x_i)])
            F[i] = result
        return F

    def marginal_icdf(self, p, dim, precision_factor=1):
        """
        Marginal inverse cumulative distribution function.

        Estimates the marginal icdf by drawing a Monte-Carlo sample.

        Parameters
        ----------
        p : array_like
            Probabilities for which the icdf is evaluated.
            Shape: 1-dimensional
        dim : int
            The dimension for which the marginal is calculated.
        precision_factor : float
            Precision factor that determines the size of the sample to draw.
            A sample is drawn of which on average precision_factor * 100
            realizations exceed the quantile. Minimum sample size is 100000.
            Defaults to 1.0

        """

        p = np.array(p)

        if self.conditional_on[dim] is None:
            # the distribution is not conditional -> it is the marginal
            return self.distributions[dim].icdf(p)
        else:
            return super().marginal_icdf(p, dim, precision_factor)

    def conditional_cdf(self, x, dim, given, *, random_state=None):
        # assert len(x) == len(given)
        distributions = self.distributions
        conditional_on = self.conditional_on

        if conditional_on[dim] is None:
            p = distributions[dim].cdf(x)
        else:
            cond_idx = conditional_on[dim]
            p = distributions[dim].cdf(x, given=given[:, cond_idx])

        return p

    def conditional_icdf(self, p, dim, given, *, random_state=None):
        # assert len(p) == len(given)
        distributions = self.distributions
        conditional_on = self.conditional_on

        if conditional_on[dim] is None:
            x = distributions[dim].icdf(p)
        else:
            cond_idx = conditional_on[dim]
            x = distributions[dim].icdf(p, given=given[:, cond_idx])
        return x

    def draw_sample(self, n, *, random_state=None):
        """
        Draw a random sample of size n.

        Parameters
        ----------
        n : int
            Sample size.
        random_state : {None, int, numpy.random.Generator}, optional
            Can be used to draw a reproducible sample.
        """

        if random_state is not None:
            # if random_state already is a np.random.Generator, default_rng returns it unaltered
            random_state = np.random.default_rng(random_state)

        samples = np.zeros((n, self.n_dim))
        for i in range(self.n_dim):
            cond_idx = self.conditional_on[i]
            dist = self.distributions[i]
            if cond_idx is None:
                samples[:, i] = dist.draw_sample(n, random_state=random_state)
            else:
                conditioning_values = samples[:, cond_idx]
                samples[:, i] = dist.draw_sample(
                    1, conditioning_values, random_state=random_state
                )

        return samples


class TransformedModel(MultivariateModel):
    def __init__(
        self,
        model: GlobalHierarchicalModel,
        transform: Callable,
        inverse: Callable,
        jacobian: Callable,
        precision_factor: float = 1.0,
        random_state: int = None,
    ):
        """A joint distribution that was defined in another variable space.

        Args:
            model (GlobalHierarchicalModel): Joint distribution in original variable space
            transform (Callable): Function to transform this model back to original variable space
            inverse (Callable): Function to transform from the original variable space to this model's space
            jacobian (Callable): jacobian matrix, see page 31 in DOI: 10.26092/elib/2181
            precision_factor (float, optional): Lower precision results in faster computation. Defaults to 1.0.
            random_state (int, optional): Can be used to fix random numbers. Defaults to None.
        """
        self.model = model
        self.transform = transform
        self.inverse = inverse
        self.jacobian = jacobian
        self.precision_factor = precision_factor
        self.random_state = random_state

        self.n_dim = self.model.n_dim
        self._sample = None

    @property
    def sample(self):
        if self._sample is None:
            self._sample = self.draw_sample(int(1e6))

        return self._sample

    def __repr__(self):
        name = "TransformedModel"
        model = repr(self.model)
        transform = repr(self.transform)
        inverse = repr(self.inverse)
        jacobian = repr(self.jacobian)

        return f"{name}(model={model}, transform={transform}, inverse={inverse}, jacobian={jacobian})"

    def fit(self, data, *args, **kwargs):
        """
        Fit joint model to data.

        Method of estimating the parameters of a probability distribution to
        given data.

        Parameters
        ----------
        data : array-like
            The data that should be used to fit the joint model.
            Shape: (number of realizations, n_dim)

        """

        return self.model.fit(self.transform(data), *args, **kwargs)

    def pdf(self, x):
        """
        Probability density function.

        Parameters
        ----------
        x : array_like
            Points at which the pdf is evaluated.
            Shape: (n, n_dim), where n is the number of points at which the
            pdf should be evaluated.

        """

        # model_pdf = self.model.pdf(self.transform(x))
        # return np.where(model_pdf >= 1e-8, model_pdf * self.jacobian(x), 0)
        return self.model.pdf(self.transform(x)) * self.jacobian(x)

    def cdf(self, x):
        """
        Cumulative distribution function.

        Parameters
        ----------
        x : array_like
            Points at which the cdf is evaluated.
            Shape: (n, n_dim), where n is the number of points at which the
            cdf should be evaluated.

        """

        x = np.atleast_2d(np.asarray_chkfinite(x))

        n_dim = self.n_dim
        integral_order = list(range(n_dim))

        def get_integral_func():
            arg_order = integral_order

            def integral_func(*args):
                assert len(args) == n_dim
                # sort arguments as expected by pdf (the models order)
                x = np.array(args)[np.argsort(arg_order)].reshape((1, n_dim))
                return self.pdf(x)

            return integral_func

        lower_integration_limits = [0] * n_dim

        integral_func = get_integral_func()

        p = np.empty(len(x))
        for i in range(len(x)):
            integration_limits = [
                (lower_integration_limits[j], x[i, j]) for j in range(n_dim)
            ]

            p[i], error = integrate.nquad(integral_func, integration_limits)

        return p

    def empirical_cdf(self, x, sample=None):
        if sample is None:
            sample = self.sample
        n = len(sample)

        x = np.atleast_2d(np.asarray_chkfinite(x))

        events = x[:, np.newaxis, :]
        leq_events = (sample <= events).all(axis=-1)
        result = leq_events.sum(axis=-1) / n

        return result

    def marginal_pdf(self, x, dim):
        """
        Marginal probability density function.

        Parameters
        ----------
        x : array_like
            Points at which the pdf is evaluated.
            Shape: 1-dimensional
        dim : int
            The dimension for which the marginal is calculated.

        """

        raise NotImplementedError()

    def marginal_cdf(self, x, dim):
        """
        Marginal cumulative distribution function.

        Parameters
        ----------
        x : array_like
            Points at which the cdf is evaluated.
            Shape: 1-dimensional
        dim : int
            The dimension for which the marginal is calculated.

        """

        raise NotImplementedError()

    def draw_sample(self, n):
        """
        Draw a random sample of size n.

        Parameters
        ----------
        n : int
            Sample size.

        """

        return self.inverse(self.model.draw_sample(n))



"""
Fit an exponentiated Weibull distribution to a significant wave height dataset.

You can compare the printed distribution parameter values to the parameter values
listed in Table 3.4 on page 45 in Haselsteiner, A. F. (2022). Offshore structures
under extreme loads: A methodology to determine design loads
[University of Bremen]. https://doi.org/10.26092/elib/1615
"""

import numpy as np
import matplotlib.pyplot as plt

from virocon import ExponentiatedWeibullDistribution, read_ec_benchmark_dataset


# Load sea state measurements from the NDBC buoy 44007.
data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_A_1year.txt")
hs = data["significant wave height (m)"].to_numpy()

# Fit the exponentiated Weibull distribution to the measurements.
dist = ExponentiatedWeibullDistribution()
dist.fit(hs)

# Print the distribution object, which outputs the estimated parameters.
print(dist)

# Plot the density of the empirical distribution and the fitted distribution.
fig, ax = plt.subplots(1, 1)
plt.hist(hs, bins=20, density=True, label="Empirical distribution")
x = np.linspace(0, max(hs), 200)
plt.plot(x, dist.pdf(x), label="Fitted distribution")
plt.xlabel("Significant wave height (m)")
plt.ylabel("Probability density (-)")
ax.legend()
# plt.savefig("virocon_exponentiated_weibull.pdf", dpi=300, bbox_inches='tight', pad_inches=0)
plt.show()
